#include <base.h>
#include <xalloc.h>
#include <lst.h>
#include <avlTree.h>
#include <print.h>

#include <memory.h>
#include <assert.h>

static avlEntry* avlEntryCreate(void){
	avlEntry* e;

	e = xalloc(sizeof(avlEntry));
	memset(e, 0, sizeof(avlEntry));
	return e;
}

static avlEntry* avlEntryInst(avlEntry* parent, void* data){
	avlEntry* a;

	a = avlEntryCreate();
	a->parent = parent;
	a->data = data;

	return a;
}

avlHead* avlCreate(const lstFns* f){
	avlHead* a;

	a = xalloc(sizeof(avlHead));
	memset(a,0,sizeof(avlHead));
	a->f = f;

	return a;
}

static void avlEntryDestroy(const lstFns* f, avlEntry* e){
	if (e->left){
		avlEntryDestroy(f, e->left);
	}
	if (e->right){
		avlEntryDestroy(f, e->right);
	}
	if (e->data){
		f->delete(e->data);
	}
	xfree(e);
}

void avlDestroy(avlHead* a){
	if (a->head){
		avlEntryDestroy(a->f, a->head);
	}
	xfree(a);
}

static avlEntry* avlEntryCopy(const lstFns* f, avlEntry* e){
	avlEntry* n;

	n = avlEntryCreate();
	n->data = f->copy(e->data);
	if (e->left){
		n->left = avlEntryCopy(f, e->left);
		n->left->parent = n;
	}
	if (e->right){
		n->right = avlEntryCopy(f, e->right);
		n->right->parent = n;
	}
	n->balance = e->balance;
	return n;
}

avlHead* avlCopy(const avlHead* a){
	avlHead* b;

	b = avlCreate(a->f);
	b->n = a->n;
	b->head = avlEntryCopy(a->f, a->head);

	return b;
}

avlEntry* avlScanForNodeFn(const avlEntry* a, const void* d, lstFnCompare f){
	lstCmp c;

top:	if (!a){
		return 0;
	}

	assert(a->data);
	c = f(d, a->data);

	switch(c){
	case lstEqual:
		return (avlEntry*)a;
	case lstLess:
		a = a->left;
		goto top;
	case lstGreater:
		a = a->right;
		goto top;
	}
	return 0;
}

avlEntry* avlScanForFn(const avlHead* a, const void* d, lstFnCompare f){
	return avlScanForNodeFn(a->head, d, f);
}

avlEntry* avlScanFor(const avlHead* a, const void* d){
	return avlScanForFn(a, d, a->f->compare);
}

boolean avlInAvl(const avlHead*a, const void* d){
	const avlEntry* e;

	e = avlScanFor(a, d);
	if (e){
		return true;
	}
	return false;
}

void* avlGetFirstData(const avlHead* h, avlEntry** ep){
	avlEntry* e;

	if (!h->head){
		*ep = 0;
		return 0;
	}

	for(e = h->head; e->left; e = e->left){
	}
	*ep = e;
	return e->data;
}

void* avlGetNextData(const avlHead* h, avlEntry** ep){
	avlEntry* e;
	USED(h);
	
	e = *ep;
	if (e->right){
		for(e = e->right; e->left; e = e->left){
		}
	} else {
		/* while we are on the right side, keep moving on up */
		while (e->parent && e != e->parent->left){
			e = e->parent;
		}
		if (e->parent){
			e = e->parent;
		} else {
			e = 0;
		}
	}

	*ep = e;
	return e ? e->data : 0;
}

void* avlGetLastData(const avlHead* h, avlEntry** ep){
	avlEntry* e;

	if (!h->head){
		*ep = 0;
		return 0;
	}

	for(e = h->head; e->right; e = e->right){
	}
	*ep = e;
	return e->data;
}

void* avlGetPrevData(const avlHead* h, avlEntry** ep){
	avlEntry* e;
	USED(h);
	
	e = *ep;
	if (e->left){
		for(e = e->left; e->right; e = e->right){
		}
	} else {
		while (e->parent && e != e->parent->right){
			e = e->parent;
		}
		if (e->parent){
			e = e->parent;
		} else {
			e = 0;
		}
	}

	*ep = e;
	return e ? e->data : 0;
}

static void avlEntryAdd(avlHead* h, avlEntry* a, void* d, const lstFns* f){
	lstCmp c;
	avlEntry* top;
	avlEntry* new_top;
	avlEntry* pivot;
	avlEntry* parent;
	avlEntry* e;

	for(top = a;;){
		c = f->compare(d, a->data);

		switch(c){
		case lstEqual:
			f->delete(a->data);
			a->data = d;
			return;
		case lstLess:
			if (a->left){
				a = a->left;
				break;
			}
			a->balance--;
			a->left = avlEntryInst(a, d);
			a->left->parent = a;
			goto done;
		case lstGreater:
			if (a->right){
				a = a->right;
				break;
			}
			a->balance++;
			a->right = avlEntryInst(a, d);
			a->right->parent = a;
			goto done;
		}
		if (a->balance){
			top = a;
		}
	}
done:
	h->n++;

	/* adjust remaining balance factors */
	for(e=a; e != top; e = parent) {
		parent = e->parent;
		if (parent->left == e){
			parent->balance--;
		} else {
			parent->balance++;
		}
	}

	switch(top->balance){
	default:
		return;
	case -2:
		if (-1 == top->left->balance) {
			new_top = top->left;
			top->left = new_top->right;
			new_top->right = top;
			new_top->balance = 0;
			top->balance = 0;
			new_top->parent = top->parent;
			top->parent = new_top;
		} else {
			pivot = top->left;
			
			new_top = pivot->right;
			pivot->right = new_top->left;
			new_top->left = pivot;
			top->left = new_top->right;
			new_top->right = top;

			if (1 == new_top->balance){
				pivot->balance = -1;
				top->balance = 0;
			} else if (-1 == new_top->balance){
				pivot->balance = 0;
				top->balance = 1;
			} else {
				pivot->balance = 0;
				top->balance = 0;
			}

			new_top->balance = 0;
			new_top->parent = top->parent;
			pivot->parent = top->parent = new_top;
			if (pivot->right){
				pivot->right->parent = pivot;
			}
		}
		if (top->left){
			top->left->parent = top;
		}
		break;
	case 2:
		if (1 == top->right->balance) {
			new_top = top->right;
			top->right = new_top->left;
			new_top->left = top;
			new_top->balance = 0;
			top->balance = 0;
			new_top->parent = top->parent;
			top->parent = new_top;
		} else {
			pivot = top->right;

			new_top = pivot->left;
			pivot->left = new_top->right;
			new_top->right = pivot;
			top->right = new_top->left;
			new_top->left = top;

			if (1 == new_top->balance){
				pivot->balance = 0;
				top->balance = -1;
			} else if (-1 == new_top->balance){
				pivot->balance = 1;
				top->balance = 0;
			} else {
				pivot->balance = 0;
				top->balance = 0;
			}
			new_top->balance = 0;
			new_top->parent = top->parent;
			pivot->parent = top->parent = new_top;
			if (pivot->left){
				pivot->left->parent = pivot;
			}
		}

		if (top->right){
			top->right->parent = top;
		}

		break;
	}

	if (!new_top->parent){
		h->head = new_top;
		return;
	}

	if (top == new_top->parent->left){ 
		new_top->parent->left = new_top;
	} else {
		new_top->parent->right = new_top;
	}
}

void avlAddEntry(avlHead* h, void*d){
	if (!h->head){
		h->head = avlEntryInst(0, d);
		h->n++;
		return;
	}

	avlEntryAdd(h, h->head, d, h->f);
}

#if 0
static int max_depth(avlEntry* e){
	avlEntry* p;
	int n;

	n = 0;
	for(p = e; p; n++){
		if (-1 == p->balance){
			p = p->left;
		} else {
			p = p->right;
		}
	}
	return n;
}
#endif

typedef enum {e_left = 0, e_right = 1} b_side;

static void _avlDeleteEntry(avlHead* h, avlEntry* stop, avlEntry* gparent, avlEntry* e, b_side bside){
	avlEntry* parent;
	avlEntry* new_top;
	avlEntry* pivot;
	avlEntry* right;
	avlEntry* left;

	/* okay, reach down and pull up stuff below this node */
	right = e->right;
	if (!right){
		/* e == gparent->side; we are collapsing e */
		if (e_left == bside){
			gparent->left = e->left;
			if (gparent->left){
				gparent->left->parent = e->parent;
			}
		} else {
			gparent->right = e->left;
			if (gparent->right){
				gparent->right->parent = e->parent;
			}
		}
	} else {
		if (!right->left){
			right->left = e->left;
			if (e_left == bside){
				gparent->left = right;
			} else {
				gparent->right = right;
			}
			right->parent = e->parent;
			if (right->left){
				right->left->parent = right;
			}
			right->balance = e->balance;
			gparent = right;
			bside = e_right;
		} else {
			for(left = right->left; left->left; left = left->left){
			}
			right = left->parent;
			right->left = left->right;
			left->left = e->left;
			left->right = e->right;
			if (e_left == bside){
				gparent->left = left;
			} else {
				gparent->right = left;
			}

			if (left->left){
				left->left->parent = left;
			}
			left->right->parent = left;
			left->parent = e->parent;

			if (right->left){
				right->left->parent = right;
			}
			left->balance = e->balance;
			gparent = right;
			bside = e_left;
		}
	}

	e->left = 0;
	e->right = 0;
	avlEntryDestroy(h->f, e);
	h->n--;

	while (gparent != stop){
		parent = gparent;
		gparent = gparent->parent;
		if (!gparent){
			gparent = stop;
		}
		
		if (e_left == bside){
			bside = gparent->left != parent;
			if (1 == ++parent->balance) {
				break;
			} else if (2 == parent->balance) {
				pivot = parent->right;
				if (-1 == pivot->balance) {
					new_top = pivot->left;
					pivot->left = new_top->right;
					new_top->right = pivot;
					parent->right = new_top->left;
					new_top->left = parent;

					if (1 == new_top->balance){
						pivot->balance = 0;
						parent->balance = -1;
					} else if (-1 == new_top->balance){
						pivot->balance = 1;
						parent->balance = 0;
					} else {
						pivot->balance = 0;
						parent->balance = 0;
					}

					new_top->balance = 0;
					new_top->parent = parent->parent;
					pivot->parent = parent->parent = new_top;
					if (pivot->left) {
						pivot->left-> parent = pivot;
					}
					if (parent->right){
						parent->right->parent = parent;
					}
					if (e_left == bside){
						gparent->left = new_top;
					} else {
						gparent->right = new_top;
					}
				} else {
					parent->right = pivot->left;
					pivot->left = parent;
					pivot->parent = parent->parent;
					parent->parent = pivot;
					if (parent->right) {
						parent->right->parent = parent;
					}

					if (e_left == bside){
						gparent->left = pivot;
					} else {
						gparent->right = pivot;
					}

					if (0 == pivot->balance) {
						pivot->balance = -1;
						parent->balance = 1;
						break;
					} else {
						pivot->balance = parent->balance = 0;
						parent = pivot;
					}
				}
			}
		} else {
			bside = gparent->left != parent;
			if (-1 == --parent->balance){
				break;
			} else if (-2 == parent->balance) {
				pivot = parent->left;
				if (1 == pivot->balance) {
					new_top = pivot->right;
					pivot->right = new_top->left;
					new_top->left = pivot;
					parent->left = new_top->right;
					new_top->right = parent;

					if (-1 == new_top->balance){
						pivot->balance = 0;
						parent->balance = 1;
					} else if (0 == new_top->balance){
						pivot->balance = 0;
						parent->balance = 0;
					} else {
						pivot->balance = -1;
						parent->balance = 0;
					}
					new_top->balance = 0;
					new_top->parent = parent->parent;
					pivot->parent = parent->parent = new_top;

					if (pivot->right){
						pivot->right->parent = pivot;
					}
					if (parent->left) {
						parent->left->parent = parent;
					}

					if (e_left == bside){
						gparent->left = new_top;
					} else {
						gparent->right = new_top;
					}
				} else {
					parent->left = pivot->right;
					pivot->right = parent;
					pivot->parent = parent->parent;
					parent->parent = pivot;
					if (parent->left) {
						parent->left->parent = parent;
					}
					if (e_left == bside){
						gparent->left = pivot;
					} else {
						gparent->right = pivot;
					}
					if (0 == pivot->balance) {
						pivot->balance = 1; 
						parent->balance = -1;
						break;
					} else {
						pivot->balance = parent->balance = 0;
						parent = pivot;
					}
				}
			}
		}
	}
}

void avlDeleteEntry(avlHead* h, avlEntry* e){
	avlEntry* gparent;
	avlEntry  bogon_flux;;
	b_side bside;

	memset(&bogon_flux, 0, sizeof(avlEntry));
	bogon_flux.left = h->head;
	bogon_flux.balance = -1;
	gparent = e->parent;
	if (!gparent){
		_avlDeleteEntry(h, &bogon_flux, &bogon_flux, e, e_left);
	} else {
		if (e == gparent->left){
			bside = e_left;
		} else {
			bside = e_right;
		}
		_avlDeleteEntry(h, &bogon_flux, gparent, e, bside);
	}
	h->head = bogon_flux.left;
}

boolean avlDeleteKey(avlHead* h, const void* d){
	avlEntry* e;

	e = avlScanForNodeFn(h->head, d, h->f->compare);
	if (e){
		avlDeleteEntry(h, e);
		return true;
	}
	return false;
}

void** avl2tab(const avlHead* a){
	int i;
	void** tab;
	void* s;
	avlEntry* e;

	if (avlZero(a)){
		return 0;
	}

	tab = xalloc(avlCount(a)*sizeof(void*));
	for(i=0, s=avlGetFirstData(a,&e); e; s=avlGetNextData(a,&e)){
		tab[i++] = s;
	}

	return tab;
}

static void pwalk(Format* f, const lstFns* fns, const avlEntry* e){
	if (!e){
		fmtprint(f, "nil");
		return;
	}

	fmtputc(f, '(');
	fmtprint(f, fns->print, e->data);
	fmtputc(f, ' ');
	pwalk(f, fns, e->left);
	fmtputc(f, ' ');
	pwalk(f, fns, e->right);
	fmtputc(f, ')');
}

fmt_action long_avl_test_conv(Format* f, const char* fmt){
	avlHead* h;
	USED(fmt);

	h = va_arg(f->args, avlHead*);
	if (!h){
		fmtappend(f, "(null)", 6);
		return FMT_verb;
	}

	//height = max_depth(h->head);
	pwalk(f, h->f, h->head);

	return FMT_verb;
}

lstHead* avlToList(avlHead* A){
	lstHead* l;
	avlEntry* e;
	void* d;

	l = lstCreate(A->f);
	for(d=avlGetFirstData(A, &e); e ; d = avlGetNextData(A, &e)){
		lstAddEntry(l, A->f->copy(d));
	}
	return l;
}

lstHead* avlToStringList(avlHead* A, const char* fmt){
	lstHead* l;
	avlEntry* e;
	void* d;

	if (!fmt){
		fmt=A->f->print;
	}

	l = lstCreate(&lstFnsString);
	for(d=avlGetFirstData(A, &e); e ; d = avlGetNextData(A, &e)){
		lstAddEntry(l, mprint(fmt, d));
	}
	return l;
}

