#include <base.h>
#include <sys/types.h>
#include <lst.h>
#include <slurp.h>
#include <pregex.h>
#include <print.h>

#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

typedef struct {
	const char* pattern;
	vlong factor;
} Utab;

static Utab binary_size_tab[] = {
	{"k(ilo)?(bytes?)?",	1<<10*1},
	{"m(ega)?(bytes?)?",	1<<10*2},
	{"g(iga)?(bytes?)?",	1<<10*3},
#ifdef GCC
	{"t(era)?(bytes?)?",	1ULL<<10*4},
	{"e(xa)?(bytes?)?",	1ULL<<10*5},
#endif
	{0,0}
};

/* ya, we're so localized */
static Utab time_tab[] = {
	{"s(econds?)?",	1		},
	{"m(inutes?)?",	60		},
	{"h(ours?)?",	3600		},
	{"d(ays?)?",	86400		},
	{"weeks?",	7*86400		},
	{"fortnights?",	14*86400	},
	{"months?",	30*86400	},	// guess at month length
	{"quarters?",	365*86400/4	},
	{"years?",	365*86400	},	// leap year problems
	{"decades?",	365*86400*10	},	// leap year problems
	{0,0}
};

static const char* match(const char* pattern, const char*s){
	static boolean once=false;
	int match_end;
	int length;
	const char* end;
	Regex* R;

	if (!once){
		once=true;
		regexInit();
	}

	R=regexCreate();
	if (!regexCompile(pattern,R)){
		fprint(2, __FILE__ ":match() can't compile static patterns.\n");
		assert(0);
	}
	length = strlen(s);
	match_end = re_match(&R->pbuf, s, length, 0, &R->regs);
	if (0 <= match_end){
		end = s+match_end;
	} else {
		end = 0;
	}

	regexDestroy(R);
	return end;
}

static boolean getFactor(const char**s, vlong* l, Utab* U){
	int i;
	const char* end;

	for(i=0; U[i].pattern; i++){
		end = match(U[i].pattern,*s);
		if (end){
			*s = end;
			*l = U[i].factor;
			return true;
		}
	}
	return false;
}

static boolean rd_parse_(const char* s, const char** endp, Utab* U, vlong* result){
	vlong l;
	vlong factor;
	vlong sign;
	const char* end;
	char operator;

	end = s;
	while (isspace(*end)){
		end++;
	}
	sign=1;
	for(;;){
		switch(*end){
		case '-':
			sign = -sign;
		case '+':
			end++;
			while (isspace(*end)){
				end++;
			}
			break;
		default:
			goto out;
		}
	}
out:
	if ('(' == *end){
		end++;
		if (!rd_parse_(end, &end, U, &l) || ')' != *end){
			*endp = end;
			return false;
		}
		end++;
	} else {
		l = strtoll(s, (char**)&end, 0);
		if (end == s){
			l=1;
		}
	}
	l *= sign;
	if (!end || !*end){
		*result = l;
		*endp = end;
		return true;
	}
	while (isspace(*end)){
		end++;
	}
	if (getFactor(&end,&factor,U)){
		l *= factor;
		while (isspace(*end)){
			end++;
		}
	}

	switch(*end){
	default:
		*endp = end;
		*result = l;
		return true;
	case ',':
	case '+':
	case '-':
	case '*':
	case '/':
		operator=*end++;
	}

	while (isspace(*end)){
		end++;
	}
	if(!rd_parse_(end, &end, U, &factor)){
		*endp = end;
		return false;
	}
	switch(operator){
	case ',':
	case '+':
		l += factor;
		break;
	case '-':
		l -= factor;
		break;
	case '*':
		l *= factor;
		break;
	case '/':
		l /= factor;
		break;
	}

	*endp = end;
	*result = l;
	return true;
}

static boolean rd_parse(const char*s, Utab* U, vlong* result){
	const char* endp;
	boolean status;

	status = rd_parse_(s,&endp,U,result);
	if (status && 0 == *endp){
		return true;
	}
	return false;
}

boolean unitTime(const char*s, vlong *result){
	return rd_parse(s,time_tab,result);
}

boolean unitBinarySize(const char*s, vlong *result){
	return rd_parse(s,binary_size_tab,result);
}
