#include <lstInternal.h>

#define internal static

internal lstEntry* lstEntryCreate(lstEntry* prev, lstEntry* next, void* data){
	lstEntry* e;

	e = xalloc(sizeof(lstEntry));
	e->next = next; if (e->next) e->next->prev = e;
	e->prev = prev; if (e->prev) e->prev->next = e;
	e->data = data;

	return e;
}

internal void lstEntryDelete(lstEntry* e){
	xfree(e);
}

internal void lstEntryDestroy(void d(void*), lstEntry* e){
	if (e->data){
		d(e->data);
	}
	xfree(e);
}

lstHead* lstCreate(const lstFns* fns){
	lstHead* l;
	
	l=xalloc(sizeof(lstHead));
	l->n = 0;
	if (fns){
		l->f = fns;
	} else {
		l->f = &lstFnsString;
	}
	l->head = 0;
	l->tail = 0;

	return l;
}

void lstDestroy(lstHead* l){
	lstEntry* e;
	lstEntry* en;

	for(e=l->head; e; e=en){
		en=e->next;
		lstEntryDestroy(l->f->delete, e);		
	}

	xfree(l);
}

lstHead* lstCopy(const lstHead* old){
	lstHead* l;
	const lstEntry* e;

	l = lstCreate(old->f);
	for(e=lstGetFirst(old); e; e = lstGetNext(e)){
		lstAddEntry(l, l->f->copy(e->data));
	}

	return l;
}

void lstAddEntry(lstHead* l, void*data){
	lstEntry* e;

	e=lstEntryCreate(l->tail, 0, data);
	l->tail = e;
	if (0 == l->head){
		l->head = e;
	}
	l->n++;
}

void lstPushEntry(lstHead* l, void*data){
	lstEntry* e;

	e=lstEntryCreate(0, l->head, data);
	l->head = e;
	if (0 == l->tail){
		l->tail = e;
	}
	l->n++;
}

void lstInsertEntry(lstHead* l, lstEntry* e, void* data){
	if (0 == e || l->tail == e){
		lstAddEntry(l,data);
		return;
	}
	lstEntryCreate(e, e->next, data);
	l->n++;	
}

void lstPushList(lstHead* l, lstHead* l2){
	lstEntry* e;
	lstEntry* h;
	lstEntry* t;
	long n;

	assert(l->f == l2->f);

	if (!lstCount(l)){
		l->head = l2->head;
		l->tail = l2->tail;
		l->n = l2->n;
		goto out;
	}

	switch(lstCount(l2)){
	case 0:
		break;
	case 1:
		e=l->head;
		h=l2->head;
		l->head = h;
		h->next = e;
		e->prev = h;
		l->n++;
		break;
	default:
		n=lstCount(l2);
		e=l->head;
		h=l2->head;
		t=l2->tail;
		l->head = h;
		t->next = e;
		e->prev = t;
		l->n+=n;
		if (l->n == n){
			l->tail = t;
		}
		break;
	}
out:	l2->n=0;
	l2->head = l2->tail = 0;
	lstDestroy(l2);
}

void lstAddList(lstHead* l, lstHead* l2){
	lstEntry* e;
	lstEntry* h;
	lstEntry* t;
	long n;

	assert(l->f == l2->f);

	if (!lstCount(l)){
		l->head = l2->head;
		l->tail = l2->tail;
		l->n = l2->n;
		goto out;
	}

	switch(lstCount(l2)){
	case 0:
		break;
	case 1:
		e=l->tail;
		h=l2->head;
		l->tail = h;
		h->prev = e;
		e->next = h;
		l->n++;
		break;
	default:
		n=lstCount(l2);
		e=l->tail;
		h=l2->head;
		t=l2->tail;
		l->tail = t;
		h->prev = e;
		e->next = h;
		l->n+=n;
		if (l->n == n){
			l->tail = t;
		}
		break;
	}
out:	l2->n=0;
	l2->head = l2->tail = 0;
	lstDestroy(l2);
}

void* lstRemoveLast(lstHead*l){
	lstEntry* e;
	void* d;

	d = 0;
	switch(lstCount(l)){
	case 0:
		break;
	case 1:
		e=l->tail;
		d=e->data;
		l->tail = l->head = 0;
		l->n = 0;
		lstEntryDelete(e);
		break;
	default:
		e=l->tail;
		d=e->data;
		l->tail = e->prev;
		l->tail->next = 0;
		l->n--;
		lstEntryDelete(e);
	}
	return d;
}

void lstDeleteLast(lstHead*l){
	void* d;

	d = lstRemoveLast(l);
	if (d){
		l->f->delete(d);
	}
}

void* lstRemoveFirst(lstHead*l){
	lstEntry* e;
	void* d;
	
	d=0;
	switch(lstCount(l)){
	case 0:
		break;
	case 1:
		e=l->head;
		d=e->data;
		l->tail = l->head = 0;
		l->n = 0;
		lstEntryDelete(e);
		break;
	default:
		e=l->head;
		d=e->data;
		l->head = e->next;
		l->head = e->next;
		l->head->prev = 0;
		l->n--;
		lstEntryDelete(e);
	}
	return d;
}

void lstDeleteFirst(lstHead*l){
	void* d;

	d = lstRemoveFirst(l);
	if (d){
		l->f->delete(d);
	}
}

// written so a forward scan will not need special processing to handle a delete.
void* lstRemoveEntry(lstHead* l, lstEntry**e){
	lstEntry* n;
	void* d;

	if (l->head == *e){
		d = lstRemoveFirst(l);
		*e = l->head;
		return d;
	}
	if (l->tail == *e){
		d = lstRemoveLast(l);
		*e = l->tail;
		return d;
	}
	assert(l->n>1);
	d = (*e)->data;
	n = (*e)->prev;
	(*e)->prev->next = (*e)->next;
	(*e)->next->prev = (*e)->prev;
	l->n--;
	lstEntryDelete(*e);
	*e = n;
	return d;
}

void lstDeleteEntry(lstHead* l, lstEntry**e){
	void* d;

	d = lstRemoveEntry(l,e);
	if (d){
		l->f->delete(d);
	}
}

lstEntry* lstGetFirst(const lstHead* l){
	return l->head;
}

lstEntry* lstGetLast(const lstHead* l){
	return l->tail;
}

lstEntry* lstGetNext(const lstEntry *e){
	return e->next;
}

lstEntry* lstGetPrev(const lstEntry *e){
	return e->prev;
}

lstEntry* lstGetCurr(const lstEntry *e){
	return (lstEntry*)e;
}

void* lstGetFirstData(const lstHead*l, lstEntry**e){
	*e = l->head;
	if (*e){
		return (*e)->data;
	}
	return 0;
}

void* lstGetNextData(const lstHead*l, lstEntry**e){
	USED(l);
	if (!*e){
		return 0;
	}
	*e = (*e)->next;
	if (*e){
		return (*e)->data;
	}
	return 0;
}

void* lstGetLastData(const lstHead*l, lstEntry**e){
	*e = l->tail;
	if (*e){
		return (*e)->data;
	}
	return 0;
}

void* lstGetPrevData(const lstHead*l, lstEntry**e){
	USED(l);
	if (!*e){
		return 0;
	}
	*e = (*e)->prev;
	if (*e){
		return (*e)->data;
	}
	return 0;
}

void* lstGetCurrData(const lstHead* l, lstEntry**e){
	USED(l);
	if (!*e){
		return 0;
	}
	return (*e)->data;
}

void** lst2tab(const lstHead*l){
	int i;
	void** tab;
	void* s;
	lstEntry* e;

	if (lstZero(l)){
		return 0;
	}

	tab = xalloc(lstCount(l)*sizeof(void*));
	for(i=0, s=lstGetFirstData(l,&e); e; s=lstGetNextData(l,&e)){
		tab[i++] = s;
	}

	return tab;
}

/* evil function */
#define CA (int(*)(const void *, const void *))
void lstSort(lstHead* l){
	void** tab;
	lstEntry *e;
	int i;

	if (lstZero(l)){
		return;
	}

	tab = lst2tab(l);
	quicksort_ptr(tab, lstCount(l), sizeof(void*), l->f->compare);

	for(i=0, e=lstGetFirst(l); e; e=lstGetNext(e)){
		e->data = tab[i++];
	}
	
	xfree(tab);
}

void** lstSortTab(lstHead* l){
	void** tab;

	if (lstZero(l)){
		lstDestroy(l);
		return 0;
	}

	tab = lst2tab(l);
	quicksort_ptr(tab, lstCount(l), sizeof(void*), l->f->compare);

	l->f = &lstFnsConstString;
	lstDestroy(l);

	return tab;
}

lstEntry* lstScanForFnE(const lstHead* l, const void* item, lstFnCompare compare, lstEntry* e){
	USED(l);
	for(; e; e=lstGetNext(e)){
		if (lstEqual == compare(e->data, item)){
			return e;
		}
	}
	return 0;
}

lstEntry* lstScanForFn(const lstHead* l, const void* item, lstFnCompare compare){
	lstEntry* e;

	for(e=lstGetFirst(l); e; e=lstGetNext(e)){
		if (lstEqual == compare(e->data, item)){
			return e;
		}
	}
	return 0;
}

lstEntry* lstScanFor(const lstHead* l, const void* item){
	return lstScanForFn(l,item, l->f->compare);
}

lstEntry* lstScanForE(const lstHead* l, const void* item, lstEntry* e){
	return lstScanForFnE(l,item, l->f->compare,e);
}

boolean lstInList(const lstHead* l, const void* item){
	if (lstScanFor(l,item)){
		return true;
	}
	return false;
}
