#include <mailhdr.h>
#include <xalloc.h>

#include <memory.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

mAddress* mAddressCreate(const char* gecos, const char* user, const char* domain, const char* comment){
	mAddress* A;

	A = xalloc(sizeof(mAddress));
	memset(A,0,sizeof(mAddress));
	A->gecos = cxcopy(gecos);
	A->user = cxcopy(user);
	if (domain){
		A->domain = lstStrstr2List(domain, strlen(domain), ".");
	}
	A->comment = cxcopy(comment);
	return A;
}

mAddress* mAddressCopy(const mAddress* A){
	mAddress* B;

	B = mAddressCreate(A->gecos, A->user, 0, A->comment);
	if (A->domain){
		B->domain = lstCopy(A->domain);
	}
	if (A->members){
		B->members = lstCopy(A->members);
	}
	return B;
}

void mAddressDestroy(mAddress* A){
	cxfree(A->gecos);
	cxfree(A->user);
	if (A->domain){
		lstDestroy(A->domain);
	}
	if (A->members){
		lstDestroy(A->members);
	}
	cxfree(A->comment);
	xfree(A);
}

lstCmp mAddressCmp(const mAddress* A, const mAddress* B){
	int n;
	const char* p;
	const char* q;
	lstEntry* pe;
	lstEntry* qe;

	n = strcasecmp(A->user, B->user);
	if (n) return n<0 ? lstLess : lstGreater;

	switch (!A->domain<<1 | !B->domain){
	case 1:
		return lstLess;
	case 2:
		return lstGreater;
	case 3:
		return lstEqual;
	}

	for(p=lstGetLastData(A->domain, &pe), q=lstGetLastData(B->domain, &qe); pe && qe; 
		p=lstGetPrevData(A->domain, &pe), q=lstGetPrevData(B->domain, &qe)){
		n = strcasecmp(p, q);
		if (n) return n<0 ? lstLess : lstGreater;
	}
	if (!qe && !pe){
		return lstEqual;
	}
	return pe ? lstGreater : lstLess;
}

void mAddress_group_int(Format* f, const mAddress* A, const char* fstring, boolean full){
	static char buf[128];

	if (full){
		if (A->user){
			//fmtprint(f, fstring, A->user);
			fmtprint(f, "%{anonymous}", long_phrase_conv, A->user);
		}

		if (lstCount(A->members)){
			snprint(buf, 128, ":\t%%/{list %%{sep %s%%{mAddress}}};", fstring);
		} else {
			snprint(buf, 128, ":;");
		}
		fmtprint(f, buf, ",\n\t\t", A->members);
	} else {
		snprint(buf, 128, "%%/{list %%{sep %s%%{mAddressBasic}}}", fstring);
		fmtprint(f, buf, " ", A->members);
	}
}

void mAddress_conv_int(Format* f, const mAddress* A, const char* fstring, boolean full){
	boolean extra;

	if (A->members){
		mAddress_group_int(f,A,fstring,full);
		return;
	}

	extra=false;
	if (full && A->gecos){
		extra=true;
	}

	if (extra){
		if (f->flags & FMT_altform) {
			fmtprint(f, "%#{phrase}", A->gecos);
		} else {
			fmtprint(f, "%{phrase}", A->gecos);
		}
		fmtappend(f, " <", 2);
	}

	fmtprint(f, fstring, A->user);

	if (!lstZero(A->domain)){
		fmtprint(f, "@%/{list}", ".", A->domain);
	}

	if (extra){
		fmtappend(f, ">", 1);
	}

	if (full && A->comment){
		fmtappend(f, " (", 2);
		fmtprint(f, "%s", A->comment);	// probablly wrong for 8bit.
		fmtappend(f, ")", 1);
	}
}

fmt_action long_mAddress_conv(Format* f, const char* fmt){
	mAddress* A;
	const char* fstring;
	USED(fmt);

	A = va_arg(f->args, mAddress*);
	if (!A){
		return FMT_verb;
	}

	fstring = "%{atom}";
	if (f->flags&FMT_altform) {
		fstring = "%{fullatom}";
	}
	mAddress_conv_int(f, A, fstring, true);
	return FMT_verb;
}

fmt_action long_mAddressBasic_conv(Format* f, const char* fmt){
	mAddress* A;
	const char* fstring;

	A = va_arg(f->args, mAddress*);
	if (!A){
		return FMT_verb;
	}
	fstring = "%{atom}";
	if (fmt){
		fstring = fmt;
	}
	mAddress_conv_int(f, A, fstring, false);
	return FMT_verb;
}

const lstFns lstFnsmAddress = {
	(lstFnCopy)mAddressCopy,
	(lstFnDelete)mAddressDestroy,
	(lstFnCompare)mAddressCmp,
	"%{mAddress}"
};

// this function is a bug.
static const char* findAddressEnd(const char*p, const char* end){
	int c;
	int level;

	for (level=0;p<end; ){
		c=*p;
		switch(c){
		case '"':
			p = scan2(p, '"', end);
			if (p<end)
				p++;
			continue;
		case '(':
			p = scancomment(0,0,p);
			if (p>end)
				p=end;			// BUG() unfinished comments will
							// core.
			continue;
		case ':':
			level++;
			break;
		case ';':
			level--;
			break;
		case '\\':
			p++;
			if (*p)
				goto done;
			break;
		case ',':
			if (0 == level)
				return p;
		}
		p++;
	}
done:
	return p;
}

static mAddress* _mAddressCreate(const char* g, const char* u, const char* d, const char* c){
	if (!g && !u && !d && !c){
		return 0;
	}
	if (!u && g){
		return mAddressCreate(u,g,d,c);
	}
	return mAddressCreate(g,u,d,c);
}

static mAddress* _mCreate(const char* g, const char* u, const char* d, const char* c){
	return _mAddressCreate(g && *g ? g : 0, u && *u ? u : 0, d && *d ? d : 0, c && *c ? c : 0);
}

mAddress* _str2mAddress(const char** S, int* N){
	mAddress* A;
	const char* s;
	int n;
	char gecos[65];
	char user[65];
	char domain[65];
	char comment[65];
	const char* gt;
	boolean have_lt;

	s = *S;
	n = *N;

	gt=findAddressEnd(s,s+n);
	*N -= gt-*S;
	*S = gt;

	comment[0] = 0;
	have_lt = false;
	s = scanphrase(gecos, 65, comment, 65, s);
	switch(*s){
	case ',':
	case 0:
	case ';':
		return _mCreate(gecos, 0, 0, 0);
	case '<':
		s++;
		have_lt = true;
	default:
		break;
	}

again:	s = scanstrws(user, 65, s);
	s = scancomment(comment, 65, s);	
	switch(*s){
	case ':':
		if (have_lt) {
			// for some reason, some use urls in leu of rfc822 addresses, 
			// notibly gnu mailman. 
			if (!strcmp(user, "mailto")){
				s++;
				goto again;
			}
		} else {
			// this is a group as in Undisclosed :;
			A = _mCreate(gecos, 0, 0, 0);
			A->members = str2mAddresses(1+s,gt-(1+s),0);
			return A;
		}
	default:
		*domain = 0;
		goto cmt;
	case '@':
		s++;
		break;
	}

	s = scanstrws(domain, 65, s);
	s = scancomment(comment, 65, s);	
cmt:	switch(*s){
	default:
		break;
	case '>':
		s++;
		break;
	}

	s = scancomment(comment, 65, s);

	A = _mCreate(gecos, user, domain, comment);

	return A;
}

lstHead* str2mAddresses(const char* s, int n, lstHead* l){
	const char* p;
	mAddress* m;

	if (!l){
		l = lstCreate(&lstFnsmAddress);
	}
	
	p = s;
	while(n){
		assert(n>0);
		m = _str2mAddress(&p, &n);
		if (!m)
			break;
		assert(n>=0);
		lstAddEntry(l, m);

		if (n && ',' == *p){
			p++, n--;
		}
		while (n && isspace(*p)){
			p++, n--;
		}
	}

	return l;
}

mAddress* str2mAddress(const char* s, int n){
	const char*p;

	p = s;
	return _str2mAddress(&p, &n);
}

