#include <mailhdr.h>
#include <xalloc.h>
#include <sys/types.h>
#include <slurp.h>
#include <rune.h>
#include <avlTree.h>

#include <memory.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>

// allow lists of numbers
ul* ulCreate(const ul const * u){
	ul* up;

	up = xalloc(sizeof(ul));
	*up = *u;
	return up;
}

void ulDestroy(ul* u){
	xfree(u);
}

lstCmp ulCmp(ul* a, ul* b){
	if (a<b){
		return lstLess;
	}
	if (a>b){
		return lstGreater;
	}
	return lstEqual;
}

fmt_action long_ul_conv(Format* f, const char* fmt){
	ul* u;
	
	u = va_arg(f->args, unsigned long*);
	if (!u){
		return FMT_verb;
	}
	fmtprint(f, fmt ? fmt : "%uld", *u);
	return FMT_verb;
}

const lstFns lstFnsUl = {
	(lstFnCopy)ulCreate,
	(lstFnDelete)ulDestroy,
	(lstFnCompare)ulCmp,
	"%{ul*}"
};

lstHead* xform_content_type(lstHead* result_hdr, const Head* src, const char* data){
	lstHead* l;
	lstEntry* e;
	char* s;
	const char* p;
	mimeString * m;
	USED(data);
	USED(result_hdr);

	l = lstCreate(&lstFnsString);
	if (!lstZero(src->values)){
		p = lstGetFirstData(src->values,&e);
		m = str2mimeString(p);
		if (!m) {
			goto lost;
		}
		s = mprint("%#{mimeString ; }", m);
		mimeStringDestroy(m);
	} else {
lost:		s = xcopy("text/plain");
	}
	lstAddEntry(l,s);
	return l;
}

lstHead* xform_copy(lstHead* result_hdr, const Head* src, const char* data){
	USED(data);
	USED(result_hdr);
	return lstCopy(src->values);
}

lstHead* xform_re(lstHead* result_hdr, const Head* src, const char* data){
	lstEntry* e;
	const char* s;
	char* m;
	boolean skip;
	USED(result_hdr);

	s=lstGetFirstData(src->values,&e);

	skip = false;
	if ('[' == data[0] && ']' == data[1]){
		skip = true;
	}

	if (*s == '[' && skip){
		goto done;
	}

	if (skip){
		data+=2;
	}

	if (strcasestr(s, data)){
		goto done;
	}
	m = mprint("%s: %s", data, s);
	lstDeleteEntry(src->values, &e);
	lstPushEntry(src->values, m);

done:
	return lstCopy(src->values);
}

lstHead* xform_split(lstHead* result_hdr, const Head* src, const char* data){
	lstHead* l;
	lstHead* split;
	lstEntry* e;
	const char* s;
	USED(result_hdr);

	l=lstCreate(&lstFnsString);
	for(s=lstGetFirstData(src->values,&e); e; s=lstGetNextData(src->values,&e)){
		split = lstStr2List(s, data);
		lstAddList(l, split);
	}
	return l;
}

static void _xform_split_email(lstHead* l, lstHead* values){
	lstEntry* e;
	const char* s;

	for(s=lstGetFirstData(values,&e); e; s=lstGetNextData(values,&e)){
		str2mAddresses(s, strlen(s), l);
	}
}

lstHead* xform_split_email(lstHead* result_hdr, const Head* src, const char* data){
	lstHead* l;
	USED(result_hdr);
	USED(data);

	l = lstCreate(&lstFnsmAddress);
	_xform_split_email(l, src->values);
	if (!lstCount(l)){
		lstDestroy(l);
		l=0;
	}
	return l;
}

lstHead* xform_append_email(lstHead* result_hdr, const Head* src, const char* data){
	lstHead* l;
	Head* h;
	USED(result_hdr);

	if (data && (h=look_head(result_hdr, data))){
		_xform_split_email(h->values, src->values);
		return 0;
	}
	l = lstCreate(&lstFnsmAddress);
	_xform_split_email(l, src->values);
	return l;
}

static Head* mksec(ul sec){
	Head* h;

	h = headCreate("X-Seconds-Hack", 0);
	lstDestroy(h->values);
	h->values = lstCreate(&lstFnsUl);
	lstAddEntry(h->values, ulCreate(&sec));
	return h;
}

lstHead* xform_date(lstHead* result_hdr, const Head* src, const char* data){
	lstEntry* e;
	lstHead* l;
	const char* s;
	ul sec;

	if (data && look_head(result_hdr, data)){
		return 0;
	}
	
	l = lstCreate(&lstFnsString);
	for(s=lstGetFirstData(src->values, &e); e; s=lstGetNextData(src->values, &e)){
		lstAddEntry(l, date2local(s, data, &sec));
		lstAddEntry(result_hdr, mksec(sec));			// for sorting
	}

	return l;
}

lstHead* xform_append(lstHead* result_hdr, const Head* src, const char* data){
	char* s;
	lstEntry* e;
	Head* h;
	lstHead* l;
	lstHead* r;
	const char* sep;

	sep = getSep(src->name);

	if (data && (h=look_head(result_hdr, data))){
		for(s = lstGetFirstData(src->values,&e); e; s = lstGetNextData(src->values,&e)){
			l = lstStrstr2List(s, strlen(s), sep);
			lstAddList(h->values, l);
		}
		return 0;
	}

	r = lstCreate(&lstFnsString);

	for(s = lstGetFirstData(src->values,&e); e; s = lstGetNextData(src->values,&e)){
		l = lstStrstr2List(s, strlen(s), sep);
		lstAddList(r, l);
	}

	return r;
}

lstHead* xform(lstHead* ls, const Xform* xform){
	lstHead* lt;
	const Head* hs;
	Head* ht;
	lstHead* l;
	const Xform* X;
	lstEntry* e;

	lt = lstCreate(&lstFnsHead);

	for(X = xform; X->from; X++){
		for (e = lstGetFirst(ls); e; e = lstGetNext(e)){
			e = look_heade(ls, X->from, e);
			if (!e){
				break;
			}
			hs = e->data;
			l = X->xform(lt, hs, X->data);
			if (!l){
				continue;
			}
			ht = look_head(lt, X->to);
			if (!ht){
				ht = headCreateList(X->to, l);
				lstAddEntry(lt, ht);
			} else {
				lstAddList(ht->values, l);
			}
		}
	}

	return lt;
}

static boolean slurpMime(bigmem*M, const char* file){
	const char* bang;
	char* f;
	boolean s;

	bang = strchr(file, '!');
	if (!bang){
		return slurp(M,file);
	}
	f = xcopyn(file, bang-file);
	s = slurp(M,f);
	xfree(f);
	return s;
}

void lstStrip(lstHead* l, avlHead* bad){
	lstEntry* e;
	const mAddress* s;

	for(s = lstGetFirstData(l, &e); e; s = lstGetNextData(l,&e)){
		if (avlScanFor(bad, s)){
			lstDeleteEntry(l, &e);
		}
	}
}

static avlHead* forList(lstHead* src){
	avlHead* a;
	Head* h;
	lstEntry* e;
	lstEntry* he;
	char* s;
	char* fer;
	char* semi;
	mAddress* try;
	boolean received;

	a = avlCreate(&lstFnsmAddress);
	received = false;

	for(h=lstGetFirstData(src, &he); he; h=lstGetNextData(src, &he)){
		if (!strcasecmp(h->name, "X-Original-To")){
			break;
		}
		if (received && !strcasecmp(h->name, "Delivered-To")){
			break;
		}
		if (strcasecmp(h->name, "Received")){
			continue;
		}

		received = true;

		for(s=lstGetFirstData(h->values,&e); e; s=lstGetNextData(h->values,&e)){
			fer = strstr(s, " for ");
			if (!fer){
				continue;
			}
			fer+=5;
			semi = strchr(fer, ';');
			if (!semi){
				continue;
			}

			try = str2mAddress(fer, semi-fer);
			avlAddEntry(a, try);
		}
	}

	return a;
}

#include <netdb.h>
#include <pwd.h>
static void addFrom(lstHead* result){
	struct passwd pw;
	struct passwd* pwp;
	char buf[1024];
	char hostname[64*3];
	char* domainname;
	struct hostent* dns;
	Head* h;

	if (look_head(result, "From")){
		return;
	}

	// fakey, fakey
	getpwuid_r(getuid(), &pw, buf, 1024, &pwp);
	if (!pwp){
		return;
	}

	if (-1 == gethostname(hostname, 64*3)){
		return;
	}

	dns = gethostbyname(hostname);
	if (!dns){
		return;
	}
	if (!(domainname = strchr(dns->h_name, '.'))){
		return;
	}
	domainname++;
	
	h = headCreate("From", 0);
	lstDestroy(h->values);
	h->values = lstCreate(&lstFnsmAddress);
;
	lstAddEntry(h->values, mAddressCreate(pwp->pw_gecos, pwp->pw_name, domainname, 0));
	lstPushEntry(result, h);
}

static avlHead* uniqueEmailTab(const char** tab, lstHead* result){
	const char** ts;
	Head* h;
	avlHead* A;
	lstEntry* e;
	mAddress* m;

	A = avlCreate(&lstFnsmAddress);
	for(ts=tab; *ts; ts++){
		h=look_head(result, *ts);
		if (!h) continue;
		for(m=lstGetFirstData(h->values, &e); e; m=lstGetNextData(h->values,&e)){
			avlAddEntry(A, mAddressCopy(m));
		}
	}
	return A;
}

static void uniqueEmail(Head* h){
	avlHead* A;
	mAddress* m;
	lstEntry* e;
	lstHead* l;

	A = avlCreate(&lstFnsmAddress);
	for(m=lstGetFirstData(h->values, &e); e; m=lstGetNextData(h->values,&e)){
		avlAddEntry(A, mAddressCopy(m));
	}
	l = avlToList(A);
	avlDestroy(A);

	lstDestroy(h->values);
	h->values = l;
}

static Head* addSendmail(lstHead* result){
	static const char* tab[] = {"To", "CC", "BCC", 0};
	Head* h;
	avlHead* A;

	A = uniqueEmailTab(tab, result);
	h=headCreate("sendmail", 0);
	lstAddEntry(h->values, mprint("sendmail %/{avl %{mAddressBasic}}\n", " ", A));
	avlDestroy(A);

	return h;
}

// just build the To line up; we'll make it unique later; later elements replace.
// eg To: fu@bar.com, target@example.com
//     From: target@example.com (fast freddy) → target@example.com (fast freddy)
static const Xform reply[] = {
	{"Reply-To",	"To",		0, "To",		xform_append_email},
	{"To",		"To",		0, "To",		xform_append_email},
	{"List-Post",	"To",		0, "To",		xform_append_email},
	{"From",		"To",		0, "To",		xform_append_email},
	{"CC",		"CC",		0, 0,		xform_split_email},
	{"References",	"References",	0, " \t\n",	xform_split},
	{"Message-Id",	"References",	0, "References",	xform_append},
	{"Message-Id",	"In-Reply-To",	0, 0,		xform_copy},
	{"Subject",	"Subject",	0, "[]Re",		xform_re},
	{0,0,0,0,0},
};

static void addMime(lstHead* l){
	Head h;
	lstEntry* e;

	memset(&h, 0, sizeof(h));
	h.name = (char*)"References";

	e = lstScanFor(l, &h);
	if (e){
		lstInsertEntry(l, e, headCreate("Content-Type", "text/plain; charset=utf-8"));
		lstInsertEntry(l, e, headCreate("Mime-Version", "1.0"));
	} else {
		lstAddEntry(l, headCreate("Mime-Version", "1.0"));
		lstAddEntry(l, headCreate("Content-Type", "text/plain; charset=utf-8"));
	}
}

void reply_mail(const char* file, const char* s, int n){
	avlHead* fer;		// keep track of who mail is for
	const Head* to;
	mimeDoc* D;
	Head* sendmail;
	USED(file);

	D=decodeBodyRef(reply, file, s, n);
	if (!D){
		return;
	}

	fer = forList(D->head_in);			// for in Received lines
	to = look_head(D->head_out, "To");
	if (to){
		lstStrip(to->values, fer);
	}
	to = look_head(D->head_out, "To");
	if (to){
//eprint("%{head}\n", to);
		uniqueEmail((Head*)to);
	}

	addFrom(D->head_out);
	to = look_head(D->head_in, "From");
	sendmail = addSendmail(D->head_out);
	addMime(D->head_out);

	print("%/{list}", "", sendmail->values);
	print("%/{list %{headquote}}\n", "", D->head_out);
	print("\n");
	if (to){
		print("%{list} writes\n\n", to->values);
	}
//	print("| %/{list}\n", "\n| ", D->lines);	// is this wrong? automatically drop subdocs?

	// this sucketh. ; alternative: writing a dualing version of %{mimeDoc}, %{mimeString}
	{
		char* body;
		SIZE_T n;
		lstHead* lines;
		lstHead* h;

		h = D->head_out; D->head_out = 0;
		body = smprint(palloc, &n, "%{mimeDoc}", D);
		D->head_out = h;
		lines = lineifyBody(body, n);
		print("| %/{list}\n", "\n| ", lines);	

		lstDestroy(lines);
		xfree(body);
	}

	mimeDocDestroy(D);	
	avlDestroy(fer);
	headDestroy(sendmail);
}

static const Xform cat[] = {
	{"Reply-To",	"Reply-To",	0, 0,	xform_copy},
	{"From",		"From",		0, 0,	xform_copy},
	{"To",		"To",		0, 0,	xform_copy},
	{"CC",		"CC",		0, 0,	xform_copy},
	{"BCC",		"BCC",		0, 0,	xform_copy},
	{"Subject",	"Subject",	0, 0,	xform_copy},
	{"Date",		"Date",		0, 0,	xform_copy},
	{0,0,0,0,0},
};

void cat_mail_int(const char* file, const char* s, int n, Xform* X){
	mimeDoc*  doc;
	USED(file);

	doc = decodeBodyRef(X, file, s, n);
	if (!doc){
		return;
	}
	print("%/{mimeDoc}\n", "", doc);
	mimeDocDestroy(doc);
}

void cat_mail(const char* file, const char* s, int n){
	cat_mail_int(file,s,n, (Xform*)cat);
}

static const Xform recat[] = {
//	{"*",	"&",	1, 0,	xform_copy},
	{"Reply-To",	"Reply-To",	0, 0,	xform_split_email},
	{"From",		"From",		0, 0,	xform_split_email},
	{"To",		"To",		0, 0,	xform_split_email},
	{"CC",		"CC",		0, 0,	xform_split_email},
	{"BCC",		"BCC",		0, 0,	xform_split_email},
	{"Subject",	"Subject",	0, 0,	xform_copy},
	{"Date",		"Date",		0, 0,	xform_copy},
	{"References",	"References",	0, 0,	xform_copy},
	{"Message-Id",	"Message-Id",	0, 0,	xform_copy},
	{"Mime-Version",	"Mime-Version",	0, 0,	xform_copy},
	{0,0,0,0,0},
};

void recat_mail(const char* file, const char* s, int n){
	mimeDoc*  doc;
	USED(file);

	doc = decodeBodyRef(recat, file, s, n);
	if (!doc){
		return;
	}

	print("%#/{mimeDoc}\n", "", doc);
	mimeDocDestroy(doc);
}

static const Xform xform_recvdchain[] = {
	{"Received",	"Received",	0, 0,		xform_copy},
	{0,0,0,0,0},
};

static fmt_action long_mSrcFrom_conv(Format* f, const char* fmt){
	mSrc* S;
	const Head* h;
	lstHead* l;
	lstEntry* e;
	lstEntry* g;
	boolean bang;
	const char*s;
	const char*p;
	USED(fmt);

	l = va_arg(f->args, lstHead*);
	if (!l){
		//fmtprint(f, "(null)");
		return FMT_verb;
	}
	bang=false;
	for(h=lstGetFirstData(l,&e); e; h=lstGetNextData(l,&e)){
	for(p=lstGetFirstData(h->values,&g); g; p = lstGetNextData(h->values,&g)){
		S = receivedstr2mSrc(p);
		if (!S)
			continue;
		if (!S->from || (!S->from->name && !S->from->value)) {
			mSrcDestroy(S);
			continue;
		}
		if (bang)
			fmtappend(f, "!", 1);
		if (!(s=S->from->name)){
			s=S->from->value;
		}
		fmtappend(f, s, strlen(s));
		xfree(S);
		bang=true;
	}
	}
	return FMT_verb;
}

void rcvdchain_mail(const char* file, const char* s, int n){
	lstHead* src;
	lstHead* result;

	src = crack_header(s, n);
	if (!src){
		return;
	}
	result = xform(src, xform_recvdchain);
	lstDestroy(src);
	print("%s: %{anonymous}\n", file, long_mSrcFrom_conv, result);
	xfree(result);
}

static const Xform xform_fwd[] = {
	{"References",	"References",	0, 0,		xform_copy},
	{"Subject",	"Subject",	0, "Fwd",		xform_re},
	{0,0,0,0,0},
};

static lstHead* tab2address(const char** addresses){
	char* m;
	lstHead* l;
	SIZE_T n;

	m = smprint(palloc, &n, "%{array}", addresses);
	l = str2mAddresses(m, n, 0);
	xfree(m);

	return l;
}

void fwd_mail(const char* file, const char* s, int n, const char** addresses){
	lstHead* src;
	lstHead* result;
	char boundary[65];
	char buf[65];
	USED(file);

	src = crack_header(s, n);
	if (!src){
		return;
	}
	result = xform(src, xform_fwd);

	addFrom(result);
	lstAddEntry(result, headCreateList("To", tab2address(addresses)));
	mimeStringBoundary(boundary, 65, "fwd");
	snprint(buf, 65, "multipart/mixed;\n\tboundary=\"%s\"", boundary);
	lstAddEntry(result, headCreate("Mime-Version", "1.0"));
	lstAddEntry(result, headCreate("Content-Type", buf));

	print("%/{list}\n", "", result);
	print("--%s\nContent-Type: text/plain; charset=utf-8\n\n", boundary);
	print("--%s\nContent-Type: message/rfc822\n\n", boundary);
	print("%#.*s\n", n, s);
	print("--%s--\n", boundary);
	
	lstDestroy(src);
	lstDestroy(result);
}

#include <fcntl.h>
static boolean detach(const char* name, const char* body, int n){
	int fd;
	int w;

	fd = open(name, O_WRONLY|O_CREAT|O_TRUNC, 0666);
	if (-1 == fd){
		eprint("cannot open <%s>: %r\n", name);
		return false;
	}

	w = safe_write_all(fd, body, n);
	close(fd);

	if (-1 == w){
		eprint("cannot write <%s>: %r\n", name);
		return false;
	}
	return true;
}

static void walk_detach(mimeDoc* doc){
	lstEntry* e;
	mimeDoc* d;

	if (doc->m && doc->m->name){
		eprint("detaching %s ...\n", doc->m->name);
		detach(doc->m->name, doc->body_out, doc->n_out);
	}
	
	if (lstZero(doc->subdocs)){
		return;
	}

	for(d=lstGetFirstData(doc->subdocs,&e); e; d=lstGetNextData(doc->subdocs,&e)){
		walk_detach(d);
	}
}

void detach_mail(const char* file, const char*s, int n){
	mimeDoc*  doc;
	USED(file);

	doc = parseBodyRef(file, s, n);
	if (!doc){
		return;
	}

	walk_detach(doc);
	mimeDocDestroy(doc);
}

static void walk_attachments(mimeDoc* doc, const char** file){
	lstEntry* e;
	mimeDoc* d;

	if (doc->m && doc->m->name){
		if (*file){
			eprint("detach %s\n", *file);
			*file = 0;
		}
		eprint("\t" "%s ... %d bytes\n", doc->m->name, doc->n_out);
		eprint("\t\t" "%>%#{mimeString}\n", doc->m);
	}
	
	if (lstZero(doc->subdocs)){
		return;
	}

	for(d=lstGetFirstData(doc->subdocs,&e); e; d=lstGetNextData(doc->subdocs,&e)){
		walk_attachments(d, file);
	}
}
void attachments_mail(const char* file, const char* s, int n){
	mimeDoc*  doc;

	doc = parseBodyRef(file, s, n);
	if (!doc){
		return;
	}

	walk_attachments(doc, &file);
	mimeDocDestroy(doc);
}
	
static void walk_mimedesc(mimeDoc* doc, const char** file, const char* address){
	lstEntry* e;
	mimeDoc* d;
	char a[128];
	int n;
			// %{mimeString ; } will suppress extra newlines.
	if (*address)	print("%s %{mimeString}\n", address, doc->m);
	else		print("%{mimeString}\n", doc->m);

	if (lstZero(doc->subdocs)){
		return;
	}

	n=0;
	for(d=lstGetFirstData(doc->subdocs,&e); e; d=lstGetNextData(doc->subdocs,&e)){
		snprint(a, 128, "%s.%02d", address, n++);
		walk_mimedesc(d, file, a);
	}
}

static void _mimedesc_mail(boolean re, const char* file, const char* s, int n){
	mimeDoc*  doc;

	if (strcmp(file, "/dev/fd/0")){
		eprint("# %s\n", file);
	}

	if (re) 	doc = decodeBodyRef(cat, file, s, n);
	else	doc = parseBodyRef(file, s, n);

	if (!doc){
		return;
	}

	walk_mimedesc(doc, &file, "");
	mimeDocDestroy(doc);
}

void mimedesc_mail(const char* file, const char* s, int n){
	_mimedesc_mail(false, file, s, n);
}

void remimedesc_mail(const char* file, const char* s, int n){
	_mimedesc_mail(true, file, s, n);
}

static const Xform from_xform[] = {
	{"From",		"From",		0, 0,		xform_split_email},
	{"Subject",	"Subject",	0, 0,		xform_copy},
	{"Date",		"Date",		0, 0,		xform_date},
	{0,0,0,0,0},
};

static const char* from_order[] = {
	"Date", 
	"From", 
	"Subject", 
	0
};

typedef struct {
	ul key;
	char* file;
	lstHead* values;
	// ↓ thread only
	boolean bhdt;		// been here, done that.
	const char* messageid;
	lstHead* references;
	lstHead* children;
} From;

extern const lstFns lstFnsFrom;

From* fromCreate(void){
	From* F;

	F = xalloc(sizeof(From));
	memset(F,0,sizeof(From));
	F->children = lstCreate(&lstFnsFrom);
	return F;
}

From* fromCopy(const From* f){
	From* F;
	const Head* h;

	F = fromCreate();
	F->file = cxcopy(f->file);
	if (f->values) {
		F->values = lstCopy(f->values);
	}
	F->messageid = look_head1(F->values, "Message-Id");
	h = look_head(F->values, "References");
	if (h){
		F->references = h->values;
	}
	if (!lstZero(f->children)) {
		lstAddList(F->children, lstCopy(f->children));
	}
	return F;
}

void fromFree(From* F){
	cxfree(F->file);
	if (F->values){
		lstDestroy(F->values);
	}
	if (F->children){
		lstDestroy(F->children);
	}
}

void fromDestroy(From* F){
	fromFree(F);
	xfree(F);
}

int from_qcmp(const void* a, const void* b){
	return ((From*)a)->key - ((From*)b)->key;
}

const lstFns lstFnsFrom = {
	(lstFnCopy)fromCopy,
	(lstFnDelete)fromDestroy,
	(lstFnCompare)from_qcmp,
	"%{head}"
};

void from_mail(From* F, const Xform* X, const char* s, int n){
	lstHead* src;
	lstHead* result;
	const Head* h;
	ul key;
	lstEntry* e;

	src = crack_header(s, n);
	if (!src){
		return;
	}
	result = xform(src, X);
	lstDestroy(src);

	key=0;
	h = look_head(result, "X-Seconds-Hack");
	if (h){
		key = *(ul*)lstGetFirstData(h->values, &e);
	}

	F->key = key;
	F->values = result;
}

// yup, this function sucks.
// and so do the column size calcuations. but what to do?
static char* chopFile(const char* file){
	const char*p;

	p = strchr(file, '_');
	if (!p){
		// postfix does this.
		p = strchr(file, '.');
		if (p) {
			p = strchr(p+1, '.');
		}
	}

	if (!p){
		return xcopy(file);
	}
	return mprint("%.*s*", p-file, file);
}

static int tab(int t, int c){
	int k;
	
	k=c+t;
	return (k/t)*t;
}

void from_print(int cols, const char** order, int n, const char* blurb, From* f){
	Head* h;
	char** stab;
	int i;
	int ts = 8;

	stab = xalloc(n*sizeof(char*));
	for(i=0; i<n; i++){
		h = look_head(f->values, order[i]);
		stab[i] = mprint("%#{head}", h);
	}

	cols -= tab(ts, strlen(blurb)) + tab(ts, 2+3+20+strlen(stab[0]));
	cols -= 2;
	if (cols<15){
		cols=15;
	}
	print("%s\t# %s"	"   %-20.20s"	"\t%-.*s\n", blurb, stab[0], stab[1], cols, stab[2]);

	for(i=0; i<n; i++){
		xfree(stab[i]);
	}
	xfree(stab);
}

#include <termios.h>
#include <fcntl.h>
#include <sys/ioctl.h>
static int windowwidth(void){
	struct winsize w;

	if (!isatty(1)){
		return 1024;
	}
	if (getenv("winid") || -1 == ioctl(1, TIOCGWINSZ, &w)){
		return 80;
	}
	return w.ws_col;
}

int frommain(int c, const char** v){
	static const char* def_v[] = {0, "mcat", 0};
	char buf[1026];
	const char* mdir;
	lstHead* mbox;
	lstEntry* e;
	const char* file;
	bigmem M;
	From* from;
	int valid_mails;
	int i;
	int ws_col;

	if (c==1){
		v = def_v;
	}

	mdir = getenv("m");
	if (!mdir){
		eprint("no $m in environment\n");
		return -1;
	}

	mbox = getDirents(mdir);
	from = 0;
	if (!lstCount(mbox)) {
		lstDestroy(mbox);
		return 0;
	}

	from = xalloc(lstCount(mbox)*sizeof(From));

	valid_mails=0;
	for(file=lstGetFirstData(mbox,&e); e; file=lstGetNextData(mbox,&e)){
		snprint(buf, 1026, "%s/%s", mdir, file);
		if (!slurp(&M, buf)){
			eprint("$m/%s: %r\n", file);
			continue;
		}
		from[valid_mails].file = chopFile(file);
		from_mail(&from[valid_mails++], from_xform, M.m, M.size);
		unslurp(&M);
	}

	qsort(from, valid_mails, sizeof(From), from_qcmp);

	ws_col = windowwidth();
	for(i=0; i<valid_mails; i++){
		snprint(buf, 1026, "%/{array %{rcquote}} %s/%s", " ", v+1, "$m", from[i].file);
		from_print(ws_col, from_order, 3, buf, &from[i]);
		xfree(from[i].file);
		lstDestroy(from[i].values);
	}

	xfree(from);
	lstDestroy(mbox);
	return 0;
}

static const Xform thread[] = {
	{"From",		"From",		0, 0,		xform_split_email},
	{"Subject",	"Subject",	0, 0,		xform_copy},
	{"Date",		"Date",		0, 0,		xform_date},
	{"References",	"References",	0, " \t\n",	xform_split},
	{"In-Reply-To",	"References",	0, 0,		xform_append},
	{"Message-Id",	"Message-Id",	0, 0,		xform_split},
	{0,0,0,0,0},
};

void _thread_from_print(Format* f, int cols, const char** order, int n, const char* blurb, From* F){
	Head* h;
	char** stab;
	int i;
	int ts = 8;
	lstEntry* e;
	From* f0;
	static int calls=0;

	calls++;
	stab = xalloc(n*sizeof(char*));
	for(i=0; i<n; i++){
		h = look_head(F->values, order[i]);
		stab[i] = mprint("%#{head}", h);
	}

	cols -= tab(ts, strlen(blurb)) + tab(ts, 2+3+20+strlen(stab[0]));
	cols -= 2;
	if (cols<15){
		cols=15;
	}
	//eprint("indent %d %d\n", f->indent, calls);
	if (f->indent){
		fmtprint(f, "%{indent     }%s%s\t# %s"	"   %-20.20s"	"\n", blurb, F->file, stab[0], stab[1], cols, stab[2]);
	} else {
		fmtprint(f, "%s%s\t# %s"	"   %-20.20s"	"\t%-.*s\n", blurb, F->file, stab[0], stab[1], cols, stab[2]);
	}

	for(i=0; i<n; i++){
		xfree(stab[i]);
	}
	xfree(stab);

	if (lstCount(F->children)){
		fmtprint(f, "%>");
		for(f0=lstGetFirstData(F->children,&e); e; f0=lstGetNextData(F->children,&e)){
			_thread_from_print(f, cols, order, n, blurb, f0);
		}
		fmtprint(f, "%<");
	}
}
void thread_from_print(int cols, const char** order, int n, const char* blurb, From* F){
	Format f;
	char buf[8*1024];

	fmt_fprint_init(1,&f, buf, sizeof buf);
	_thread_from_print(&f, cols, order, n, blurb, F);
	fmt_fprint_close(&f);
}

static const From*
findMessageID(lstHead* l, lstHead* refs){
	lstEntry* e;
	const From* f;
	const From* c;
	lstEntry* re;
	const char* r;

	for(f=lstGetFirstData(l,&e); e; f=lstGetNextData(l,&e)){
#if 1
		if (f->children){
			c = findMessageID(f->children, refs);
			if (c) {
				//eprint("found child\n");
				return c;
			}
		}
#endif
		if (!f->messageid)
			continue;
		for(r=lstGetFirstData(refs,&re); re; r=lstGetNextData(refs,&re)){
			if (0 == strcmp(r, f->messageid)){
				// eprint("found %s\n", r);
				return f;
			}
		}
	}
	return 0;
}

static lstHead*
thread_mail(From* tab, int n){
	const From* end;
	From* p;
	lstHead* l;
	const From* c;
	lstHead* root;

	l = lstCreate(&lstFnsFrom);
	end = tab+n;
	for(p = tab; p<end; p++){
		if (!p->references){
			lstAddEntry(l, fromCopy(p));
			continue;
		}

		for(root=l; ; root=c->children){
			c = findMessageID(root, p->references);
			if (!c || !c->children){
				break;
			}
		}
		lstAddEntry(root, fromCopy(p));
	}

	return l;
}

int threadmain(int c, const char** v){
	static const char* def_v[] = {0, "mcat", 0};
	char buf[1026];
	char* mdir;
	lstHead* mbox;
	lstEntry* e;
	const char* file;
	bigmem M;
	From* from;
	int valid_mails;
	int ws_col;
	lstHead* l;
	Head* h;
	From* fp;

	if (c==1){
		v = def_v;
	}

	mdir = getenv("m");
	if (!mdir){
		eprint("no $m in environment\n");
		return -1;
	}

	mbox = getDirents(mdir);
	from = 0;
	if (!lstCount(mbox)) {
		lstDestroy(mbox);
		return 0;
	}

	from = xalloc(lstCount(mbox)*sizeof(From));
	memset(from,0,lstCount(mbox)*sizeof(From));

	valid_mails=0;
	for(file=lstGetFirstData(mbox,&e); e; file=lstGetNextData(mbox,&e)){
		snprint(buf, 1026, "%s/%s", mdir, file);
		if (!slurp(&M, buf)){
			eprint("$m/%s: %r\n", file);
			continue;
		}
		from[valid_mails].file = chopFile(file);
		from_mail(&from[valid_mails], thread, M.m, M.size);
		unslurp(&M);
		from[valid_mails].messageid = look_head1(from[valid_mails].values, "Message-Id");
		h = look_head(from[valid_mails].values, "References");
		if (h){
			from[valid_mails].references = h->values;
		}
		valid_mails++;
	}

	qsort(from, valid_mails, sizeof(From), from_qcmp);
	l = thread_mail(from, valid_mails);

	ws_col = windowwidth();
	for(fp = lstGetFirstData(l, &e); e ; fp = lstGetNextData(l, &e)){
		snprint(buf, 1026, "%/{array %{rcquote}} %s/", " ", v+1, "$m");
		thread_from_print(ws_col, from_order, 3, buf, fp);
	}

	lstDestroy(l);

	for(int i=0; i<valid_mails; i++){
		fromFree(&from[i]);
	}

	xfree(from);
	lstDestroy(mbox);
	return 0;
}

static int fwdmain(int c, const char**v, void (*f)(const char*, const char*, int, const char**)){
	bigmem M;

	v++; c--;

	switch(c){
	case 0:
		break;
	case 1:
		if (!slurpFd(&M, 0)){
			eprint("%s: %r\n", "<stdin>");
			exit(1);
		}
		f("/dev/fd/0", M.m, M.size, v+1);
		unslurp(&M);
		break;
	default:
		if (!slurp(&M, *v)){
			eprint("%s: %r\n", *v);
			exit(1);
		}
		f(*v, M.m, M.size, v+1);
		unslurp(&M);
	}
	return 0;
}

void xheader(const Xform* X, const char* fname, const char* s, int n){
	From* F;
	USED(fname);
	
	F = fromCreate();
	from_mail(F,X,s,n);
	print("%/{list %{head  }}", "", F->values);
	fromDestroy(F);
}

// HELP:
// doesn't allow you to specify how to refrobnicate each header line
// and References and Content-Type are get special treatment
static int xheadermain(int c, const char**v, void (*f)(const Xform*, const char*, const char*, int)){
	bigmem M;
	lstHead* l;
	lstEntry* e;
	Xform* X;
	Xform* x;
	const char* s;
	int size;

	v++; c--;
	if (!c){
		return 0;
	}

	l = lstStrstr2List(*v, strlen(*v), ":");
	if (!l){
		return 0;
	}
	size = sizeof(Xform)*(1+lstCount(l));
	X = xalloc(size);
	memset(X,0, size);
	x=X;
	for(s=lstGetFirstData(l,&e); e; x++, s=lstGetNextData(l,&e)){
		x->to = s;
		x->from = s;
		if (0 == strcasecmp(s,"References")){
			x->data = " \t\n";
			x->xform = xform_split;
		} else if (0 == strcasecmp(s, "Content-Type")){
			x->xform = xform_content_type;
		} else if (isAddressHdr(s)){
			x->xform = xform_split_email;
		} else {
			x->xform = xform_copy;
		}
	}

	v++; c--;
	switch(c){
	case 0:
		if (!slurpFd(&M, 0)){
			eprint("%s: %r\n", "<stdin>");
			exit(1);
		}
		f(X, "/dev/fd/0", M.m, M.size);
		unslurp(&M);
		break;
	default:
		for(; *v; v++){
			if (!slurpMime(&M, *v)){
				eprint("%s: %r\n", *v);
				exit(1);
			}
			f(X, *v, M.m, M.size);
			unslurp(&M);
		}
	}

	xfree(X);
	lstDestroy(l);

	return 0;
}

const char* posix_sucks_basename(const char* file){
	const char* p;

	p = strrchr(file, '/');
	if (p){
		return p+1;
	}
	return file;
}

int main(int c, const char** v){
	bigmem M;
	const char* base;
	void (*f)(const char*, const char*, int);

	mailprintInstall();

	f = cat_mail;
	base = posix_sucks_basename(*v);
	if (0 == strcmp(base, "mcat")){
		f=cat_mail;
	} else if (0 == strcmp(base, "mrecat")){
		f=recat_mail;
	} else if (0 == strcmp(base, "reply")){
		f=reply_mail;
	} else if (0 == strcmp(base, "detach")){
		f=detach_mail;
	} else if (0 == strcmp(base, "attachments")){
		f=attachments_mail;
	} else if (0 == strcmp(base, "mimedesc")){
		f=mimedesc_mail;
	} else if (0 == strcmp(base, "remimedesc")){	// ≡ mrecat | mimedesc
		f=remimedesc_mail;
	} else if (0 == strcmp(base, "thread")){
		return threadmain(c,v);
	} else if (0 == strcmp(base, "from")){
		return frommain(c,v);
	} else if (0 == strcmp(base, "fwd")){
		return fwdmain(c,v, fwd_mail);
	} else if (0 == strcmp(base, "xheader")){
		return xheadermain(c, v, xheader);
	} else if (0 == strcmp(base, "rcvdchain")){
		f=rcvdchain_mail;
	}

	if (1 == c){
		if (!slurpFd(&M, 0)){
			eprint("%s: %r\n", "<stdin>");
			exit(1);
		}
		f("/dev/fd/0", M.m, M.size);
		unslurp(&M);
	} else {
		for(v++; *v; v++){
			if (!slurpMime(&M, *v)){
				eprint("%s: %r\n", *v);
				exit(1);
			}
			f(*v, M.m, M.size);
			unslurp(&M);
		}
	}
	exit(0);
	return 0;
}
