#include <mailhdr.h>
#include <xalloc.h>

#include <memory.h>
#include <string.h>

mimeDoc* mimeDocCreate(void){
	mimeDoc* D;

	D = xalloc(sizeof(mimeDoc));
	memset(D,0,sizeof(mimeDoc));
	return D;
}

mimeDoc* mimeDocCopy(const mimeDoc* A){
	mimeDoc* B;

	B = mimeDocCreate();
	if (A->m) {
		B->m = mimeStringCopy(A->m);
	}

	if (A->head_in) {
		B->head_in = lstCopy(A->head_in);
	}
	if (A->head_out) {
		B->head_out = lstCopy(A->head_out);
	}

	if (A->body_allocated){
		B->body_allocated = true;
		B->body_in = xcopyn(A->body_in, A->n_in);
	} else {
		B->body_in = A->body_in;
	}
	B->n_in = A->n_in;

	if (A->subdocs){
		B->subdocs = lstCopy(A->subdocs);
	}

	if (A->body_out){
		B->body_out = xcopyn(A->body_out, A->n_out);
		B->n_out = A->n_out;
	}

	if (A->lines){
		B->lines = lstCopy(A->lines);
	}

	return B;
}

void mimeDocDestroy(mimeDoc* D){
	if (D->m) {
		mimeStringDestroy(D->m);
	}
	if (D->head_in) {
		lstDestroy(D->head_in);
	}
	if (D->head_out) {
		lstDestroy(D->head_out);
	}
	if (D->body_allocated){
		cxfree(D->body_in);
	}
	if (D->subdocs) {
		lstDestroy(D->subdocs);
	}
	if (D->out_allocated){
		cxfree(D->body_out);
	}
	if (D->lines) {
		lstDestroy(D->lines);
	}
	xfree(D);
}

void mimeDocSetBody(mimeDoc* D, char* s, int n, boolean allocated){
	if (D->out_allocated){
		cxfree(D->body_out);
	}
	D->body_out = s;
	D->n_out = n;
	D->out_allocated = allocated;
}

static boolean printable(const mimeString* m){
	static const char* bad[] = {"image", "video", "application", 0};
	const char** v;
	const char* p;
	ul n;

	if (!m){
		return true;	// probablly plain rfc-822.
	}

	if (!m->type){
		return true;
	}

	p = strchr(m->type, '/');
	if (!p){
		return true;
	}

	n = p - m->type;
	for(v=bad; *v; v++){
		if (n == strlen(*v) && 0 == strncmp(*v, m->type, n)){
			return false;
		}
	}
	return true;
}

// suitable for mailing
static void mimeDoc_altconv_int(Format* f, const char* fmt, mimeDoc* D, int d){
	lstEntry* e;
	lstHead* l;
	mimeDoc* subD;
	const char* boundary;

	l = D->head_out;
	if (!l){
		l = D->head_in;
	}

	fmtprint(f, "%/{list %{headquote}}", "", l);
	if (D->m){
		if (0 == d && !look_head(l, "Mime-Version")) {
			fmtprint(f, "Mime-Version: 1.0\n");
		}
		if (!look_head(l, "Content-Type")) {
			fmtprint(f, "%{mimeString}\n", D->m);
		}
		if (!look_head(l, "Content-Disposition")) {
			fmtprint(f, "%{mimeStringDispo}\n", D->m);
		}
	}

	// boundary is nil for message/rfc822.
	if (D->subdocs){
		boundary=0;
		if (D->m){
			boundary = D->m->boundary;
		}

		for(subD=lstGetFirstData(D->subdocs, &e); e; subD=lstGetNextData(D->subdocs,&e)){
			if (boundary) 	fmtprint(f, "\n--%s\n", boundary);
			//else 		fmtprint(f, "\n");
			mimeDoc_altconv_int(f,fmt,subD,d+1);
		}
		if (boundary) fmtprint(f, "\n--%s--\n", boundary);
		return;
	}

	if (!printable(D->m)){
		fmtprint(f, "Content-Transfer-Encoding: base64\n");
		fmtappend(f, "\n", 1);
		fmtprint(f, "%#.*{base64}", D->n_out, D->body_out);
		return;
	}
	fmtappend(f, "\n", 1);
	fmtappend(f, D->body_out, D->n_out);
}


// suitable for viewing
static void mimeDoc_conv_int(Format* f, const char* fmt, mimeDoc* D){
	lstEntry* e;
	mimeDoc* subD;
	const char* boundary;

	fmtprint(f, "%/{list}", "", D->head_out);

	if (D->subdocs){
		boundary = 0;
		if (D->m && lstCount(D->subdocs)>1){
			boundary = D->m->boundary;
		}
		for(subD=lstGetFirstData(D->subdocs, &e); e; subD=lstGetNextData(D->subdocs,&e)){
			if (boundary) 	fmtprint(f, "\n--%s\n", boundary);
			//else 		fmtprint(f, "\n");
			mimeDoc_conv_int(f,fmt,subD);
		}
		//if (boundary) fmtprint(f, "\n--%s--\n\n", D->m->boundary);
		return;
	}

	if (!printable(D->m)){
		fmtprint(f, "\n[elided]\n");
		return;
	}

	if (D->lines){
		fmtprint(f, "\n%/{list}\n", fmt ? fmt : "\n", D->lines);
		return;
	}
	fmtappend(f, "\n", 1);
	fmtappend(f, D->body_out, D->n_out);
}

fmt_action long_mimeDoc_conv(Format* f, const char* fmt){
	mimeDoc* D;
	USED(fmt);

	D=va_arg(f->args, mimeDoc*);
	if (!D){
		return FMT_verb;
	}

	if (f->flags & FMT_altform){
		mimeDoc_altconv_int(f, fmt, D, 0);
	} else {
		mimeDoc_conv_int(f, fmt, D);
	}

	return FMT_verb;
}

const lstFns lstFnsMimeDoc = {
	(lstFnCopy)mimeDocCopy,
	(lstFnDelete)mimeDocDestroy,
	0,
	"%{mimeDoc}"
};

