// basics for parsing Received lines with a view to as-yet-unimplemented spam filtering.
#include <mailhdr.h>
#include <xalloc.h>

#include <stdlib.h>
#include <memory.h>
#include <ctype.h>
#include <assert.h>

static const char* mSrcTypeStr[(int)mSrcInfinite+1] = {
	0,
	"Received",
	"Reply-To",
	"Sender",
	"From",
	0
};

mSrc* mSrcCreate(void){
	mSrc* R;

	R = xalloc(sizeof(mSrc));
	memset(R,0,sizeof(mSrc));
	R->datetime = -1;
	return R;
}

static nv* nvCCopy(const nv* n){
	if (!n){
		return 0;
	}
	return nvCopy(n);
}

static mAddress* mAddressCCopy(const mAddress* m){
	if (m){
		return mAddressCopy(m);
	}
	return 0;
}

mSrc* mSrcInst(mSrcType type, nv* from, nv* by, char* via, lstHead* with, char* msgid, mAddress* _for, long t, long o){
	mSrc* R;

	R = mSrcCreate();
	R->type = type;
	R->from = from;
	R->by = by;
	R->via = via;
	R->with = with;
	R->msgid = msgid;
	R->_for = _for;
	R->datetime = t;
	R->offset = o;
	return R;
}

mSrc* mSrcCopy(const mSrc* R){
	return mSrcInst(R->type, nvCCopy(R->from), nvCCopy(R->by), cxcopy(R->via), lstCCopy(R->with), cxcopy(R->msgid), mAddressCCopy(R->_for), R->datetime, R->offset);
}

static void nvCDestroy(nv* n){
	if (n){
		nvDestroy(n);
	}
}

void mSrcDestroy(mSrc* R){
	nvCDestroy(R->from);
	nvCDestroy(R->by);
	cxfree(R->via);
	if (R->with) lstDestroy(R->with);
	cxfree(R->msgid);
	if (R->_for) mAddressDestroy(R->_for);
}

static void fmtnv(Format* f, const char* fmt, const nv* n){
	boolean sp;

	sp=false;
	if (n->name){
		fmtprint(f, fmt, n->name);
		sp=true;
	}
	if (n->value){
		if (sp) fmtappend(f," ",1);
		fmtprint(f, "[%s]", n->value);
	}
}

fmt_action long_mSrc_conv(Format* f, const char* fmt){
	mSrc* R;
	boolean sp;
	char b[50];

	R = va_arg(f->args, mSrc*);
	if (!R){
		fmtprint(f, "(null)");
		return FMT_verb;
	}
	if (!fmt){
		fmt = "%{atom}";
		if (f->flags&FMT_altform){
			fmt = "%{fullatom}";
		}
	}

	if (f->flags&FMT_altform){
		fmtprint(f, "%s: ", mSrcTypeStr[(int)R->type]);
	}

	switch(R->type){
	case mSrcReceived:
		sp=false;
		if (R->from){
			fmtprint(f, "from ");
			fmtnv(f, "%s", R->from);
			sp=true;
		}

		if (R->by){
			if (sp) fmtappend(f, " ", 1);
			fmtprint(f, "by ");
			fmtnv(f, "%s", R->by);
			sp=true;
		}

		if (R->via){
			if (sp) fmtappend(f, " ", 1);
			fmtprint(f, "via ");
			fmtprint(f, fmt, R->via);
			sp=true;
		}

		if (!lstZero(R->with)){
			if (sp) fmtappend(f, " ", 1);
			fmtprint(f, "with ");
			snprint(b, 50, "%%/{list %s}", fmt);
			fmtprint(f, b, " ", R->via);
			sp=true;
		}
	
		if (R->msgid){
			if (sp) fmtappend(f, " ", 1);
			fmtprint(f, "msgid ");
			fmtprint(f, fmt, R->msgid);
			sp=true;
		}

		if (R->_for){
			if (sp) fmtappend(f, " ", 1);
			fmtprint(f, "for ");
			snprint(b, 50, "%%{mAddress %s}", fmt);
			fmtprint(f, b, R->_for);
			sp=true;
		}
		if (-1 != R->datetime){
			fmtprint(f, "; ");
			mailtime(f, R->datetime, R->offset);
			sp=true;
		}
		break;
	case mSrcReplyTo:
	case mSrcSender:
	case mSrcFrom:
		snprint(b, 50, "%%{mAddress %s}", fmt);
		fmtprint(f, b, R->_for);
		break;
	default:
		assert(0 == "bad mSrcType in long_mSrc_conv");
	}
	return FMT_verb;
}

const lstFns lstFnsMSrc = {
	(lstFnCopy)mSrcCopy,
	(lstFnDelete)mSrcDestroy,
	0,
	"%{mSrc}"
};

static const char* keywords[] = {
	";",
	"by",
	"for",
	"from",
	"id",
	"via",
	"with",
	0
};
	
static boolean iskeyword(const char* s){
	const char** k;

	for(k = keywords; *k; k++){
		if (!strcasecmp(s,*k)){
			return true;
		}
	}
	return false;
}


static const char* scanhost(const char* s, nv** np){
	const char* p;
	char c[65];
	char n[65];
	char v[65];

	*np = 0;
	*n = 0;
	*v = 0;

	s = scancomment(c, 65, s);
	if (*s != '['){
		p = scanstrws(n, 65, s);
		if (!iskeyword(n)){
			s=p;
		} else {
			*n=0;
		}
	}
	s = scancomment(c, 65, s);
	if (*s == '['){
		s = scansqbracket(v, 65, s);
	}

	if (*n || *v){
		*np = nvInst(*n ? xcopy(n) : 0, *v ? xcopy(v) : 0);
	}
	return s;
}

mSrc* receivedstr2mSrc(const char* s){
	char label[65];
	char tag[65];
	mSrc* m;
	const char* p;
	int n;

	m = mSrcCreate();
	m->type = mSrcReceived;

	s = scanstrws(label, 65, s);
	if (!strcasecmp(label, "from")){
		s = scanhost(s, &m->from);
		s = scanstrws(label, 65, s);
	}

	if (!strcasecmp(label, "by")){
		s = scanhost(s, &m->by);
		s = scanstrws(label, 65, s);
	}

	if (!strcasecmp(label, "via")){
		s = scanstrws(tag, 65, s);
		m->via = xcopy(tag);
		s = scanstrws(label, 65, s);
	}

	if (!strcasecmp(label, "with")){
		m->with = lstCreate(&lstFnsString);
		do {
			s = scanstrws(label, 65, s);
			if (iskeyword(label))
				break;
			lstAddEntry(m->with, xcopy(label));
		} while (*s && *s != ';');
	}

	if (!strcasecmp(label, "id")){
		s = scanstrws(tag, 65, s);
		m->msgid = xcopy(tag);
		s = scanstrws(label, 65, s);
	}

	if (!strcasecmp(label, "for")){
		p = strchr(s, ';');
		if (p){
			n=p-s;
		} else {
			n = strlen(s);
		}
		m->_for = _str2mAddress(&s, &n);
	}

	if (';' == *s){
		s++;
		while (isspace(*s))
			s++;
		s = scandate(s, 0, &m->datetime, &m->offset);
	}

	return m;
}
