/*
**
**  Copyright:		Copyright © public domain
**
**  RCS Info:		$Date: 2005/09/23 22:56:05 $ $Revision: 1.10 $
**
**			$Id: print.c,v 1.10 2005/09/23 22:56:05 quanstro Exp quanstro $
**
**  Filename:		$Source: /home/quanstro/src/elocate/print/RCS/print.c,v $
**
**  Originator:		Erik Quanstrom
**			202 328 3688
**			quanstro@quanstro.net
**
**  Organization:	Erik Quanstrom
**			1455 Chapin ST NW
**			Washington DC 20009
**			Ph: 202 328 3688	FAX: zippidy
**
**  Description:
**
**	This is the libprint public domain library, a replacement for the
**	assorted standard I/O printing functions
**
**  Limitations:
**
**	1.  %g (borrowed from plan9) doesn't appear to pad correctly.
**	2.  %L and %A are not supported.
**	3.  %{} doesn't allow private format tabs.
**	4.  dflt_fmttab modifications are not protected by a lock.
**	5.  %λ single-character unicode formats don't work. %{λ} however does.
**
**  Modification History:
**
** 	14-nov-1995 quanstro Created.
**	23-feb-1998 quanstro updated to v1.1
**	05-nov-1999 sward    Added anonymous format support.
**	29-may-2005 quanstro added c99 long long support.
**	30-may-2005 quanstro added %g, %b, and the flags +, ' '
**	17-jun-2005 quanstro added %hh; tab is '\t'.
**	01-jul-2005 quanstro %s now respects utf-8; anonymous formats on the way out.
**	21-jul-2005 quanstro anonymous formats killed; %X.
**	28-oct-2005 quanstro reset client_data like f1 and f2 on new format.
**
******************************************************************************/
/* print.c -- formatted printing routines */
/* this by Scott Schwartz, derived from Paul Haahr's version of 12/91 */

#include <printInternal.h>
#include <rune.h>
#include <hash.h>
#include <xalloc.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>

#if defined ( __GNUC__ ) && !defined(LINUX)
extern void *memcpy (void *, const void *, long unsigned int);
#endif

#ifdef AIX
extern int errno;
#endif

/*
 * functions for inserting strings in the format buffer
 */

int fmtappend(Format *format, const char *s, SIZE_T len) {
	format->reqlen += len;
	if (format->error)
		return 0;
	while (len > (SIZE_T)(format->bufend - format->buf)) {
		const SIZE_T split = format->bufend - format->buf;
		memcpy (format->buf, s, split);
		format->buf += split;
		s += split;
		len -= split;
		if (format->grow(format, len))
	    		return 0;
    	}
    	memcpy(format->buf, s, len);
    	format->buf += len;
    	return 1;
}

int fmtputr(Format *f, const Rune r){
	int n;
	char s[UTFmax];

	n=runetochar(s,&r);
	return fmtappend(f, s, n);
}

int fmtputc(Format *f, char c) {
	return fmtappend(f, &c, 1);
}

int fmtcat(Format *format, const char *s) {
	return fmtappend(format, s, strlen(s));
}

/*
 * conversion functions
 *	true return -> flag changes only, not a conversion
 */

#define Flag(name, flag) \
static fmt_action name(Format *format, int c) { \
	USED(c) ; \
	format->flags |= flag; \
	return FMT_flag; \
}

Flag(spconv,	FMT_space)
Flag(uconv,	FMT_unsigned)
Flag(hconv,	format->flags & FMT_short ? FMT_byte : FMT_short)
Flag(lconv,	format->flags & FMT_long  ? FMT_vlong : FMT_long)
Flag(altconv,	FMT_altform)
Flag(leftconv,	FMT_leftside)
Flag(signconv,	FMT_sign)
Flag(dotconv,	FMT_f2set)
Flag(commaconv,	FMT_comma)

static fmt_action digitconv(Format *format, int c) {
	if (format->flags & FMT_f2set) {
		format->f2 = 10 * format->f2 + c - '0';
	} else {
		format->flags |= FMT_f1set;
		format->f1 = 10 * format->f1 + c - '0';
	}
	return FMT_flag;
}

static fmt_action zeroconv(Format *format, int c) {
	USED(c);

	if (format->flags & (FMT_f1set | FMT_f2set)) {
		return digitconv(format, '0');
	} else if(format->flags & FMT_zeropad) {
		return digitconv(format, '0');
	}
	format->flags |= FMT_zeropad;
	return FMT_flag;
}

static fmt_action starconv (Format *format, int c) {
	int n = va_arg(format->args, int);
	USED(c);

	if (format->flags & FMT_f2set) {
		format->f2 = n;
	} else {
		format->flags |= FMT_f1set;
		format->f1 = n;
	}
	return FMT_flag;
}

void fmtpad(Format *format, SIZE_T len, char c) {
	do; while (len-- != 0 && fmtputc(format, c));
}

static fmt_action cconv(Format *format, int c) {
	USED(c);
	fmtputc(format, va_arg(format->args, int));
	return FMT_verb;
}

static fmt_action litconv(Format *format, int c) {
	fmtputc(format, c);
	return FMT_verb;
}

fmt_action fmt_badconv (Format *format, int c) {
	eprint("libprint: bad conversion char '%c'\n", c);
	format->error = 1; /* XXX - stop or not? consume an argv? */
	return FMT_verb;
}

/* write the current write length into an int* argument */
static fmt_action nconv(Format* format, int c){
	int *p;
	USED(c);

	p = va_arg(format->args, int*);
	if (p){
		*p = format->buf - format->bufbegin + format->flushed;
	}
	return FMT_verb;
}

/* do we have to store errno in the format structure because
   the grow proc could make a syscall that sets errno before
   we get here? 
*/
static fmt_action rconv(Format *format, int c) {
	int e;
	USED(c);

	e = errno;
	if (format->client_data) {
		e = (int)format->client_data;
	}

	if (format->flags & FMT_altform) {
		fmtprint(format, "%d", e);
	} else {
		fmtcat(format, strerror(e));
	}
	return FMT_verb;
}

static fmt_action delim_convf(Format *f, int c){
	USED(c);

	f->client_data = va_arg(f->args, void*);
	return FMT_flag;
}

static fmt_action glconv(Format *f, int c){
	if (c == '>') {
		f->indent++;
	} else if (c == '<'){
		f->indent--;
		if (0>=f->indent){
			f->indent = 0;
		}
	}
	return FMT_verb;
}

static fmt_action nlconv(Format* f, int c){
	USED(c);
	fmtappend(f, "\n", 1);
	fmtpad(f, f->indent, '\t');
	return FMT_verb;
}

/*
 * conversion table management
 */

static Fmtconv dflt_fmttab[FMT_MAXCONV];

static Fmtconv* ensure_inittab (Fmtconv fmttab[]) {
	int i;

	if (fmttab[0] != 0) {
		return fmttab;
	}

	for (i = 0; i < FMT_MAXCONV; i++) {
		fmttab[i] = fmt_badconv;
	}

	fmttab[' '] = spconv;
	fmttab['%'] = litconv;

	fmttab['s'] = sconv;
	fmttab['c'] = cconv;

	fmttab['d'] = dconv;
	fmttab['i'] = dconv;
	fmttab['o'] = oconv;
	fmttab['x'] = xconv;
	fmttab['b'] = bconv;
	fmttab['r'] = rconv;
	fmttab['n'] = nconv;
	fmttab['X'] = Xconv;

	fmttab['u'] = uconv;
	fmttab['h'] = hconv;
	fmttab['l'] = lconv;
	fmttab['#'] = altconv;
	fmttab['-'] = leftconv;
	fmttab['+'] = signconv;
	fmttab['.'] = dotconv;
	fmttab[','] = commaconv;
	fmttab['*'] = starconv;
//	fmttab['q'] = quote_conv;		c.f. stdprint: long format %{rcquote}
	fmttab['C'] = Cconv;
	fmttab['S'] = Sconv;
	fmttab['/'] = delim_convf;
	fmttab['e'] = __efgfmt;
	fmttab['E'] = __efgfmt;
	fmttab['f'] = __efgfmt;
	fmttab['F'] = __efgfmt;
	fmttab['g'] = __efgfmt;
	fmttab['G'] = __efgfmt;
	fmttab['>'] = glconv;
	fmttab['<'] = glconv;
	fmttab['\n'] = nlconv;
	fmttab['0'] = zeroconv;
	for (i = '1'; i <= '9'; i++){
		fmttab[i] = digitconv;
	}

	return fmttab;
}

Fmtconv fmtinstall(int c, Fmtconv f) {
	Fmtconv oldf;

	ensure_inittab(dflt_fmttab);
	c &= FMT_MAXCONV - 1;
	oldf = dflt_fmttab[c];
	if (f) {
		dflt_fmttab[c] = f;
	}
	return oldf;
}

/* --- */

static struct {
	char* name;
	SIZE_T len;
	unsigned int hash;
	Fmtlongconv f;
} dflt_longfmttab[FMT_MAXLONGCONV];

static void long_fmtdeinstall(void){
	int i;

	for (i=0; dflt_longfmttab[i].name && i<FMT_MAXLONGCONV; i++){
		xfree(dflt_longfmttab[i].name);
		dflt_longfmttab[i].name = 0;
		dflt_longfmttab[i].len = 0;
		dflt_longfmttab[i].hash = 0;
		dflt_longfmttab[i].f = 0;
	}
}

Fmtlongconv long_fmtinstall(const char* s, Fmtlongconv f){
	static int been_here = 0;
	Fmtlongconv oldf;
	SIZE_T i, len, hash;

	if (!been_here){
		atexit(long_fmtdeinstall);
		been_here = 1;
	}

	len = strlen(s);
	hash = full_name_hash((unsigned char*)s, len);

	for (i=0; dflt_longfmttab[i].name && i<FMT_MAXLONGCONV; i++){
		if (len == dflt_longfmttab[i].len && hash == dflt_longfmttab[i].hash && 0 == strcmp(dflt_longfmttab[i].name, s)){
			oldf = dflt_longfmttab[i].f;
			dflt_longfmttab[i].f = f;
			return oldf;
		}
	}
	dflt_longfmttab[i].name = xcopy(s);
	dflt_longfmttab[i].len = len;
	dflt_longfmttab[i].hash = hash;
	dflt_longfmttab[i].f = f;
	return 0;
}

/* --- */

static void fmtinitbuf(Format *format, char *buf, SIZE_T size, int (*grow)(Format *, SIZE_T)) {
	format->buf	  = buf;
	format->bufbegin  = buf;
	format->bufend    = buf + size;
	format->grow	  = grow;
	format->flushed   = 0;
	format->error	  = 0;
	format->reqlen    = 0;
	format->indent    = 0;
}

static void fmtinitarg(Format *format, va_list args, Fmtconv* fmttab, void* client_data) {
	format->args		= args;
	format->fmttab		= 0;
	format->client_data 	= client_data;
}
		     
/*
 * the formatting engine
 */

static fmt_action fancy_badconv(Format* f, const char*a, const char* b){
	eprint("libprint: %s %s\n", a, b);
	f->error = 1; /* XXX - stop or not? consume an argv? */
	return FMT_verb;
}

static const char* scan2matching(const char* s,int up, int down){
	int level;
	int c;

	level = 1;
	for (c = *s;; c = *++s) {
		if (c == up){
			level++;
		} else if (c == down){
			level--;
			if (0 == level){
				return s;
			}
		} else if (0 == c){
			break;
		}
	};
	return 0;
}
		
static fmt_action fancyfmt(Format*f, const char**s, int c){
	const char* end;
	SIZE_T len;
	int i;
	unsigned int hash;
	char* arguments;
	char* t;
	fmt_action rval;

	end = scan2matching(*s, '{', '}');
	if (0 == end){
		return fancy_badconv(f,"badly formed %{}.", "no '}' found.");
	}

	len = end-*s;
	if (0 == len){
		return fancy_badconv(f,"badly formed %{}.", "no string found.");
	}

	hash = init_name_hash();
	t = (char*)*s;
	c = *(const unsigned char *)t;
	do {
		hash = partial_name_hash(c, hash);
		t++;
		c = *(const unsigned char *)t;
	} while (t<end && (c != ' '));
	hash = end_name_hash(hash);

	arguments = 0;
	if (t && t<end){
		len=t-*s;
		if (1 < end-t){
			arguments = xcopyn(t+1,end-t-1);
		}
	}

	for (i=0; dflt_longfmttab[i].name && i<FMT_MAXLONGCONV; i++){
		if (len != dflt_longfmttab[i].len) continue;
		if (hash != dflt_longfmttab[i].hash) continue;
		if (0 == strncmp(dflt_longfmttab[i].name, *s, len)){
			*s = end+1;
			rval = dflt_longfmttab[i].f(f,arguments);
			if (arguments){
				xfree(arguments);
			}
			return rval;
		}
	}

	if (arguments){
		xfree(arguments);
	}
	eprint("libprint: in %s. unknown format '%#.*s'.\n", "%{}", end-*s, *s);
	*s=end+1;
	f->error = 1; /* XXX - stop or not? consume an argv? */
	return FMT_verb;
}

int fmtengine(Format * f, const char *fmt) {
	const unsigned char *s;
	Fmtconv* fmttab;
	int c;

	s = (const unsigned char*)fmt;
	fmttab = f->fmttab;

	if (0 == fmttab) {
		fmttab = ensure_inittab(dflt_fmttab);
	}

	for (;;) {
		c = *s++;

		switch (c) {
		case '%':
			f->flags = f->f1 = f->f2 = 0;
			f->client_data = 0;
			do c = *s++; while (FMT_flag == (c == '{' ? fancyfmt(f, (const char**)&s,c) : fmttab[c](f, c)));
			break;
		case '\0':
			return f->buf - f->bufbegin + f->flushed;
		default:
			fmtputc(f, c);
			break;
		}
	}
}


/*
 * the public entry points
 */

int fmtprint (Format *format, const char *fmt,...) {
	int n = -format->flushed;
	va_list ap, saveargs;
	int saveflags;
	void* saveclient_data;

	va_start (ap, fmt);
	saveargs = format->args;
	saveflags = format->flags;
	saveclient_data = format->client_data;
	format->args = ap;
	n += fmtengine (format, fmt);
	va_end (format->args);
	format->args = saveargs;
	format->flags = saveflags;
	format->client_data = saveclient_data;

	return n + format->flushed;
}

/* --- */

static int writeall(int fd, char *buf, SIZE_T n) {
	int i, remain;

	for (i = 0, remain = n; remain > 0; buf += i, remain -= i) {
		if ((i = write(fd, buf, remain)) <= 0) {
			if ((errno == EAGAIN) 
#ifdef EWOULDBLOCK
				|| (errno == EWOULDBLOCK)
#endif
				) {
				continue;
			} else {
				break;
			}
		}
	}

	return n - remain; 
}

/* --- */

int fmt_fprint_flush (Format *format, SIZE_T more) {
	SIZE_T n = format->buf - format->bufbegin;
	SIZE_T r;
	USED(more);

	r = writeall(format->u.fd, format->bufbegin, n);
	format->flushed += r;
	format->buf -= r;
	format->error |= (r != n);

	return format->error;
}

int fmt_fprint_init(int fd, Format* format, char* buf, SIZE_T buf_size){
	fmtinitbuf(format, buf, buf_size, fmt_fprint_flush);
	fmtinitarg(format, 0, 0, 0);
	format->u.fd = fd;
	return 1;
}

int fmt_fprint_close(Format* format){
	fmt_fprint_flush(format, 0);
	return format->error ? -1 : format->flushed;
}

int vfprint(int fd, const char *fmt, va_list ap) {
	char buf[2048];
	Format format;

	fmtinitbuf(&format, buf, sizeof buf, fmt_fprint_flush);
	fmtinitarg(&format, ap, 0, 0);
	format.u.fd = fd;

	fmtengine(&format, fmt);
	fmt_fprint_flush(&format, 0);
	return format.error ? -1 : format.flushed;
}

int fprint(int fd, const char *fmt,...) {
	int n;
	va_list ap;

	va_start(ap, fmt);
	n = vfprint(fd, fmt, ap);
	va_end(ap);
	return n;
}

int print(const char *fmt,...) {
	int n;
	va_list ap;

	va_start(ap, fmt);
	n = vfprint(1, fmt, ap);
	va_end(ap);
	return n;
}

int eprint(const char *fmt,...) {
	int n;
	va_list ap;

	va_start(ap, fmt);
	n = vfprint(2, fmt, ap);
	va_end(ap);
	return n;
}

/* --- */

int fmt_memprint_grow (Format *format, SIZE_T more) {
	char *buf = 0;
	SIZE_T len = format->bufend - format->bufbegin + 1;
	SIZE_T buflen = format->buf - format->bufbegin;
	len = (len >= more)
		? len * 2 
		: ((len + more) + FMT_PRINT_ALLOCSIZE) &~ (FMT_PRINT_ALLOCSIZE - 1);

	buf = format->u.p(format->bufbegin, len);
	if (buf) {
		format->buf		 = buf + buflen;
		format->bufbegin = buf;
		format->bufend   = buf + len - 1;
		return 0;
	}
	return format->error = 1;
}

char *fmt_memprint(Format *format, const char *fmt, SIZE_T* len) {
	char* buf;
	int n;

	buf = format->u.p(0, FMT_PRINT_ALLOCSIZE);
	if (!buf) {
		format->error = 1;
		return 0;
	}

	fmtinitbuf(format, buf, FMT_PRINT_ALLOCSIZE - 1, fmt_memprint_grow);
	n = fmtengine(format, fmt);
	*format->buf = '\0';
	if (len) {
		*len = n;
	}

	return format->error ? 0 : format->bufbegin;
}

char *vsmprint(Alloc a, SIZE_T* len, const char* fmt, va_list ap) {
	Format format;

	fmtinitarg(&format, ap, 0, 0);
	format.u.p = a;

	return fmt_memprint(&format, fmt, len);
}

char *smprint(Alloc alloc, SIZE_T* len, const char* fmt, ...) {
	char* result;
	va_list ap;

	va_start(ap, fmt);
	result = vsmprint(alloc, len, fmt, ap);
	va_end(ap);
	return result;
}

/* --- */

int fmt_snprint_grow (Format *format, SIZE_T more) {
    USED(more);
    return format->error=1;
}

int vsnprint (char *buf, int buflen, const char *fmt, va_list ap) {
	Format format;

	fmtinitbuf(&format, buf, buflen - 1, fmt_snprint_grow);
	fmtinitarg(&format, ap, 0, 0);
	format.args = ap;

	fmtengine (&format, fmt);
	*format.buf = '\0';
	return format.reqlen;
}

int snprint (char *buf, int buflen, const char *fmt,...) {
	int n;
	va_list ap;

	va_start(ap, fmt);
	n = vsnprint(buf, buflen, fmt, ap);
	va_end(ap);
	return n;
}

int sprint (char *buf, const char *fmt,...) {
	int n;
	va_list ap;

	va_start(ap, fmt);
	n = vsnprint(buf, FMT_SPRINT_BUFSIZ, fmt, ap);
	va_end(ap);
	return n;
}
